"""
Created on Sun Aug 11 22:30:06 2019.

@author: tageldim

"""
import copy
import warnings
from io import BytesIO

import numpy as np
from PIL import Image, ImageDraw
from shapely.geometry.polygon import Polygon

Image.MAX_IMAGE_PIXELS = None


def get_image_from_htk_response(resp):
    """Given a histomicsTK girder response, get np array image.

    Parameters
    ----------
    resp : object
           response from server request

    Returns
    -------
    Pillow Image object
        a pillow Image object of the image

    """
    image_content = BytesIO(resp.content)
    image_content.seek(0)
    image = Image.open(image_content)
    image = image.convert('RGB')
    return np.uint8(image)


def scale_slide_annotations(slide_annotations, sf):
    """Scales up or down annotations in a slide.

    Works in place, but returns slide_annotations anyways.

    Parameters
    ----------
    slide_annotations : list of dicts
        response from server request

    sf : float
        scale factor to multiply coordinates

    Returns
    -------
    list of dicts

    """
    if sf == 1.0:
        return slide_annotations

    for _annidx, ann in enumerate(slide_annotations):
        for element in ann['annotation']['elements']:
            for key in element.keys():

                if key in ['height', 'width']:
                    element[key] = int(element[key] * sf)

                elif key in ['center', 'points']:
                    element[key] = (np.array(element[key]) * sf).astype(int).tolist()
    return slide_annotations


def get_scale_factor_and_appendStr(gc, slide_id, MPP=None, MAG=None):
    """Get how much is request region smaller than base.

    This also gets the string to append to server request for getting rgb.

    Parameters
    ----------
    gc : Girder client instance
        gc should be authoenticated.
    slide_id : str
        girder id of slide
    MPP : float or None
        Microns-per-pixel -- best use this as it's more well-defined than
        magnification which is more scanner/manufacturer specific.
        MPP of 0.25 often roughly translates to 40x
    MAG : float or None
        If you prefer to use whatever magnification is reported in slide.
        If neither MPP or MAG is provided, everything is retrieved without
        scaling at base (scan) magnification.

    Returns
    -------
    float
        how much smaller (0.1 means 10x smaller) is requested image
        compared to scan magnification (slide coordinates)
    str
        string to appnd to server request for getting slide region

    """
    slide_info = gc.get('/item/%s/tiles' % slide_id)

    if (MPP is not None) and (slide_info['mm_x'] is not None):
        mm = 0.001 * MPP
        sf = slide_info['mm_x'] / mm
        appendStr = '&mm_x=%.8f&mm_y=%.8f' % (mm, mm)

    elif (MAG is not None) and (slide_info['magnification'] is not None):
        sf = MAG / slide_info['magnification']
        appendStr = '&magnification=%.8f' % MAG

    else:
        warnings.warn(
            'NO SLIDE MAGNIFICATION FOUND; BASE MAGNIFICATION USED!',
            RuntimeWarning, stacklevel=1)
        sf = 1.0
        appendStr = ''

    return sf, appendStr


def rotate_point_list(point_list, rotation, center=(0, 0)):
    """Rotate a certain point list around a central point. Modified from.

    javascript version at: https://github.com/girder/large_image/blob/master/
    ... web_client/annotations/rotate.js.

    Parameters
    ----------
    point_list : list of tuples
        (x,y) coordinates

    rotation : float
        degrees (in radians)

    center : tuple of ints
        central point coordinates

    Returns
    -------
    list of tuples

    """
    point_list_rotated = []

    for point in point_list:

        cos = np.cos(rotation)
        sin = np.sin(rotation)
        x = point[0] - center[0]
        y = point[1] - center[1]

        point_list_rotated.append((
            int(x * cos - y * sin + center[0]),
            int(x * sin + y * cos + center[1])))

    return point_list_rotated


def get_rotated_rectangular_coords(
        roi_center, roi_width, roi_height, roi_rotation=0):
    """Given data on rectangular ROI center/width/height/rotation.

    Get the unrotated abounding box coordinates around rotated ROI. This
    of course is applicable to any rotated rectangular annotation.

    Parameters
    ----------
    roi_center : tuple or list
        (x, y) format
    roi_width : float
    roi_height : float
    roi_rotation : float

    Returns
    -------
    dict
        includes roi corners (x, y) and bounds

    """
    # Get false bounds
    x_min = roi_center[0] - int(roi_width / 2)
    x_max = roi_center[0] + int(roi_width / 2)
    y_min = roi_center[1] - int(roi_height / 2)
    y_max = roi_center[1] + int(roi_height / 2)

    # Get pseudo-corners of rectangle (before rotation)
    roi_corners_false = [
        (x_min, y_min), (x_max, y_min),
        (x_max, y_max), (x_min, y_max)]

    # Get *true* coordinates of rectangle corners
    roi_corners = rotate_point_list(roi_corners_false,
                                    rotation=roi_rotation,
                                    center=roi_center)
    roi_corners = np.array(roi_corners, dtype=np.int32)

    # pack into dict
    roi_info = {
        'roi_corners': roi_corners,
        'x_min': roi_corners[:, 0].min(),
        'x_max': roi_corners[:, 0].max(),
        'y_min': roi_corners[:, 1].min(),
        'y_max': roi_corners[:, 1].max(),
    }

    return roi_info


def get_bboxes_from_slide_annotations(slide_annotations):
    """Given a slide annotation list, gets information on bounding boxes.

    Parameters
    ----------
    slide_annotations : list of dicts
        response from server request

    Returns
    -------
    Pandas DataFrame
        The columns annidx and elementidx encode the
        dict index of annotation document and element, respectively, in the
        original slide_annotations list of dictionaries

    """
    from pandas import DataFrame

    element_infos = DataFrame(columns=[
        'annidx', 'elementidx', 'type', 'group',
        'xmin', 'xmax', 'ymin', 'ymax'])

    for annidx, ann in enumerate(slide_annotations):
        for elementidx, element in enumerate(ann['annotation']['elements']):

            elno = element_infos.shape[0]
            element_infos.loc[elno, 'annidx'] = annidx
            element_infos.loc[elno, 'elementidx'] = elementidx
            element_infos.loc[elno, 'type'] = element['type']

            # get bounds
            if element['type'] == 'polyline':
                coords = np.array(element['points'])[:, :-1]
                xmin, ymin = (int(j) for j in np.min(coords, axis=0))
                xmax, ymax = (int(j) for j in np.max(coords, axis=0))

            elif element['type'] == 'rectangle':
                roiinfo = get_rotated_rectangular_coords(
                    roi_center=element['center'],
                    roi_width=element['width'],
                    roi_height=element['height'],
                    roi_rotation=element['rotation'])
                xmin, ymin = roiinfo['x_min'], roiinfo['y_min']
                xmax, ymax = roiinfo['x_max'], roiinfo['y_max']

            else:
                continue

            # add group or infer from label
            if 'group' in element.keys():
                element_infos.loc[elno, 'group'] = element['group']
            elif 'label' in element.keys():
                element_infos.loc[elno, 'group'] = element['label']['value']

            element_infos.loc[elno, 'xmin'] = xmin
            element_infos.loc[elno, 'xmax'] = xmax
            element_infos.loc[elno, 'ymin'] = ymin
            element_infos.loc[elno, 'ymax'] = ymax
            element_infos.loc[elno, 'bbox_area'] = int(
                (ymax - ymin) * (xmax - xmin))

    return element_infos


def _get_coords_from_element(element):

    # get bounds
    if element['type'] == 'polyline':
        coords = np.int32(element['points'])[:, :-1]

    elif element['type'] == 'rectangle':
        roiinfo = get_rotated_rectangular_coords(
            roi_center=element['center'],
            roi_width=element['width'],
            roi_height=element['height'],
            roi_rotation=element['rotation'])
        coords = roiinfo['roi_corners']  # for rotated rectangles
        if element['rotation'] != 0:
            element['type'] = 'polyline'

    elif element['type'] == 'point':
        xmin = xmax = int(element['center'][0])
        ymin = ymax = int(element['center'][1])
        coords = np.array(
            [(xmin, ymin), (xmax, ymin), (xmax, ymax),
             (xmin, ymax), (xmin, ymin)], dtype='int32')

    else:
        msg = 'Unsupported element type:'
        raise Exception(msg, element['type'])

    return coords


def _maybe_crop_polygon(vertices, bounds_polygon):
    """Crop bounds to desired area using shapely polygons."""
    all_vertices = []

    # First, we get the polygon or polygons which result from
    # intersection. Keep in mind that a particular annotation may
    # "ecit then enter" the cropping ROI, so it may be split into
    # two or more polygons by the splitting process. That's which
    # this method's input is one set of vertices, but after cropping
    # it may return one or more sets of vertices
    try:
        elpoly = Polygon(vertices).buffer(0)
        polygon = bounds_polygon.intersection(elpoly)

        if polygon.geom_type in ('Polygon', 'LineString'):
            polygons_to_add = [polygon]
        else:
            polygons_to_add = [
                p for p in polygon if p.geom_type == 'Polygon']

    except Exception as e:
        # if weird shapely errors -->
        # ignore this polygon collection altogether
        print(e.__repr__())
        return all_vertices

    # Second, once we have the shapely polygons, assuming no errors,
    # we parse into usable coordinates that we can add to the table
    for poly in polygons_to_add:
        try:
            if polygon.area > 2:
                all_vertices.append(np.array(
                    poly.exterior.xy, dtype=np.int32).T)
        except Exception as e:
            # weird shapely errors --> ignore this polygon
            print(e.__repr__())

    return all_vertices


def _parse_coords_to_str(vertices):
    return (
        ','.join(str(j) for j in vertices[:, 0]),
        ','.join(str(j) for j in vertices[:, 1]))


def _add_element_to_final_df(vertices, cfg):
    """Add a single element to the final dataframe.

    Note that we wrap this into a method so that when (if) we split an
    annotation element into multiple polygons, we can add these
    separately.
    """
    elno = cfg.element_infos.shape[0]

    # Add element information to element dataframe
    cfg.element_infos.loc[elno, 'annidx'] = cfg.annidx
    cfg.element_infos.loc[elno, 'annotation_girder_id'] = cfg.ann['_id']
    cfg.element_infos.loc[elno, 'elementidx'] = cfg.elementidx
    cfg.element_infos.loc[elno, 'element_girder_id'] = cfg.element['id']
    cfg.element_infos.loc[elno, 'color'] = str(cfg.element['lineColor'])

    # Now we can add offset to ensure coordinates are relative to the
    # cropping bounds (i.e. they would correspond to an RGB image
    # of the same region and could be used to create a mask or
    # to encode object boundaries etc
    if (cfg.cropping_bounds is not None) or (
            cfg.cropping_polygon_vertices is not None):
        vertices[:, 0] = vertices[:, 0] - cfg.x_shift
        vertices[:, 1] = vertices[:, 1] - cfg.y_shift

    # get bounds for this polygon. Remember, these may have been
    # changed after it was cropped using shapely
    xmin, ymin = np.min(vertices, axis=0)
    xmax, ymax = np.max(vertices, axis=0)

    # parse to string for inclusion in pd dataframe
    x_coords, y_coords = _parse_coords_to_str(vertices)

    # add group or infer from label
    if 'group' in cfg.element.keys():
        cfg.element_infos.loc[elno, 'group'] = str(cfg.element['group'])
    elif 'label' in cfg.element.keys():
        cfg.element_infos.loc[elno, 'group'] = str(
            cfg.element['label']['value'])

    # add label or infer from group
    if 'label' in cfg.element.keys():
        cfg.element_infos.loc[elno, 'label'] = str(
            cfg.element['label']['value'])
    elif 'group' in cfg.element.keys():
        cfg.element_infos.loc[elno, 'label'] = cfg.element_infos.loc[
            elno, 'group']

    cfg.element_infos.loc[elno, 'type'] = str(cfg.element['type'])
    cfg.element_infos.loc[elno, 'xmin'] = int(xmin)
    cfg.element_infos.loc[elno, 'xmax'] = int(xmax)
    cfg.element_infos.loc[elno, 'ymin'] = int(ymin)
    cfg.element_infos.loc[elno, 'ymax'] = int(ymax)
    cfg.element_infos.loc[elno, 'bbox_area'] = int(
        (ymax - ymin) * (xmax - xmin))
    cfg.element_infos.loc[elno, 'coords_x'] = x_coords
    cfg.element_infos.loc[elno, 'coords_y'] = y_coords


def parse_slide_annotations_into_tables(
        slide_annotations, cropping_bounds=None,
        cropping_polygon_vertices=None, use_shapely=False):
    """Given a slide annotation list, parse into convenient tabular format.

    If the annotation is a point, then it is just treated as if it is a
    rectangle with zero area (i.e. xmin=xmax). Rotated rectangles are treated
    as polygons for simplicity.

    Parameters
    ----------
    slide_annotations : list of dicts
        response from server request

    cropping_bounds : dict or None
        if given, must have keys XMIN, XMAX, YMIN, YMAX. These are the
        bounds to which the polygons may be cropped using shapely,
        if the param use_shapely is True. Otherwise, the polygon
        coordinates are just shifted relative to these bounds without
        actually cropping.

    cropping_polygon_vertices : nd array or None
        if given, is an (m, 2) nd array of vertices to crop bounds.
        if the param use_shapely is True. Otherwise, the polygon
        coordinates are just shifted relative to these bounds without
        actually cropping.

    use_shapely : bool
        see cropping_bounds description.

    Returns
    -------
    Pandas DataFrame
        Summary of key properties of the annotation documents. It has the
        following columns:
        - annotation_girder_id
        - _modelType
        - _version
        - itemId
        - created
        - creatorId
        - public
        - updated
        - updatedId
        - groups
        - element_count
        - element_details

    Pandas DataFrame

        The individual annotation elements (polygons, points, rectangles).
        The columns annidx and elementidx encode the dict index of annotation
        document and element, respectively, in the original slide_annotations
        list of dictionaries. It has the following columns:

        - annidx
        - annotation_girder_id
        - elementidx
        - element_girder_id
        - type
        - group
        - label
        - color
        - xmin
        - xmax
        - ymin
        - ymax
        - bbox_area
        - coords_x
        - coords_y

    """
    from pandas import DataFrame

    # we use this object to pass params to split method into sub-methods
    # and avoid annoying linting ("method too complex") issue
    class Cfg:
        def __init__(self):
            pass
    cfg = Cfg()
    cfg.cropping_bounds = cropping_bounds
    cfg.cropping_polygon_vertices = cropping_polygon_vertices
    cfg.use_shapely = use_shapely

    cfg.annotation_infos = DataFrame(columns=[
        'annotation_girder_id', '_modelType', '_version',
        'itemId', 'created', 'creatorId',
        'public', 'updated', 'updatedId',
        'groups', 'element_count', 'element_details',
    ])

    cfg.element_infos = DataFrame(columns=[
        'annidx', 'annotation_girder_id',
        'elementidx', 'element_girder_id',
        'type', 'group', 'label', 'color',
        'xmin', 'xmax', 'ymin', 'ymax', 'bbox_area',
        'coords_x', 'coords_y',
    ])

    if cfg.cropping_bounds is not None:
        assert cfg.cropping_polygon_vertices is None, \
            'either give cropping bounds or vertices, not both'
        xmin, xmax, ymin, ymax = (
            cfg.cropping_bounds['XMIN'], cfg.cropping_bounds['XMAX'],
            cfg.cropping_bounds['YMIN'], cfg.cropping_bounds['YMAX'])
        bounds_polygon = Polygon(np.array([
            (xmin, ymin), (xmax, ymin),
            (xmax, ymax), (xmin, ymax), (xmin, ymin),
        ], dtype='int32'))
        cfg.x_shift = xmin
        cfg.y_shift = ymin

    elif cfg.cropping_polygon_vertices is not None:
        bounds_polygon = Polygon(np.int32(cfg.cropping_polygon_vertices))
        cfg.x_shift, cfg.y_shift = np.min(cfg.cropping_polygon_vertices, 0)

    # go through annotation elements and add as needed
    for cfg.annidx, cfg.ann in enumerate(slide_annotations):

        annno = cfg.annotation_infos.shape[0]

        # Add annotation document info to annotations dataframe

        cfg.annotation_infos.loc[annno, 'annotation_girder_id'] = cfg.ann['_id']

        for key in [
                '_modelType', '_version',
                'itemId', 'created', 'creatorId',
                'public', 'updated', 'updatedId']:
            if key in cfg.ann:
                cfg.annotation_infos.loc[annno, key] = cfg.ann[key]

        if 'groups' in cfg.ann:
            cfg.annotation_infos.loc[annno, 'groups'] = str(cfg.ann['groups'])

        if '_elementQuery' in cfg.ann:
            cfg.annotation_infos.loc[annno, 'element_count'] = cfg.ann[
                '_elementQuery']['count']
            cfg.annotation_infos.loc[annno, 'element_details'] = cfg.ann[
                '_elementQuery']['details']

        for idx, element in enumerate(cfg.ann['annotation']['elements']):
            cfg.elementidx = idx
            cfg.element = element
            coords = _get_coords_from_element(copy.deepcopy(element))

            # crop using shapely to desired bounds if needed
            # IMPORTANT: everything till this point needs to be
            # relative to the whole slide image
            if ((cfg.cropping_bounds is None) and
                    (cfg.cropping_polygon_vertices is None)) \
                    or (element['type'] == 'point') \
                    or (not use_shapely):
                all_coords = [coords]
            else:
                all_coords = _maybe_crop_polygon(coords, bounds_polygon)

            # now add polygons one by one
            for vertices in all_coords:
                _add_element_to_final_df(vertices, cfg=cfg)

    return cfg.annotation_infos, cfg.element_infos


def np_vec_no_jit_iou(bboxes1, bboxes2):
    """Fast, vectorized IoU.

    Source: https://medium.com/@venuktan/vectorized-intersection-over-union ...
            -iou-in-numpy-and-tensor-flow-4fa16231b63d

    Parameters
    ----------
    bboxes1 : np array
        columns encode bounding box corners xmin, ymin, xmax, ymax
    bboxes2 : np array
        same as bboxes 1

    Returns
    -------
    np array
        IoU values for each pair from bboxes1 & bboxes2

    """
    x11, y11, x12, y12 = np.split(bboxes1, 4, axis=1)
    x21, y21, x22, y22 = np.split(bboxes2, 4, axis=1)
    xA = np.maximum(x11, np.transpose(x21))
    yA = np.maximum(y11, np.transpose(y21))
    xB = np.minimum(x12, np.transpose(x22))
    yB = np.minimum(y12, np.transpose(y22))
    interArea = np.maximum((xB - xA + 1), 0) * np.maximum((yB - yA + 1), 0)
    boxAArea = (x12 - x11 + 1) * (y12 - y11 + 1)
    boxBArea = (x22 - x21 + 1) * (y22 - y21 + 1)
    iou = interArea / (boxAArea + np.transpose(boxBArea) - interArea)
    return iou


def _get_idxs_for_all_rois(GTCodes, element_infos):
    """Get indices of ROIs within the element_infos dataframe.

    (Internal)

    """
    roi_labels = list(GTCodes.loc[GTCodes.loc[:, 'is_roi'] == 1, 'group'])
    idxs_for_all_rois = []
    for idx, elinfo in element_infos.iterrows():
        if elinfo.group in roi_labels:
            idxs_for_all_rois.append(idx)
    return idxs_for_all_rois


def get_idxs_for_annots_overlapping_roi_by_bbox(
        element_infos, idx_for_roi, iou_thresh=0.0):
    """Find indices of **potentially** included annotations within the ROI.

    We say "potentially" because this uses the IoU of roi and annotation as a
    fast indicator of potential inclusion. This helps dramatically scale down
    the number of annotations to look through. Later on, a detailed look at
    whether the annotation polygons actually overlap the ROI can be done.

    Parameters
    ----------
    element_infos : pandas DataFrame
        result from running get_bboxes_from_slide_annotations()
    idx_for_roi : int
        index for roi annotation within the element_infos DF
    iou_thresh : float
        overlap threshold to be considered within ROI

    Returns
    -------
    list
        indices relative to element_infos

    """
    bboxes = np.array(
        element_infos.loc[:, ['xmin', 'ymin', 'xmax', 'ymax']],
        dtype='float')
    iou = np_vec_no_jit_iou(bboxes[idx_for_roi, :][None, ...], bboxes2=bboxes)
    iou = np.concatenate((np.arange(iou.shape[1])[None, ...], iou))
    iou = iou[:, iou[1, :] > iou_thresh].astype(int)

    overlaps = set(iou[0, :].tolist()) - {idx_for_roi}

    return list(overlaps)


def create_mask_from_coords(coords):
    """Create a binary mask from given vertices coordinates.

    Source: This is modified from code by Juan Carlos from David Gutman Lab.

    Parameters
    ----------
    coords : np array
        must be in the form (e.g. ([x1,y1],[x2,y2],[x3,y3],.....,[xn,yn])),
        where xn and yn corresponds to the nth vertex coordinate.

    Returns
    -------
    np array
        binary mask

    """
    polygon = coords.copy()

    # use the smallest bounding region, calculated from vertices
    min_x, min_y = np.min(polygon, axis=0)
    max_x, max_y = np.max(polygon, axis=0)
    # get the new width and height
    width = int(max_x - min_x)
    height = int(max_y - min_y)
    # shift all vertices to account for location of the smallest bounding box
    polygon[:, 0] = polygon[:, 0] - min_x
    polygon[:, 1] = polygon[:, 1] - min_y

    # convert to tuple form for masking function (nd array does not work)
    vertices_tuple = tuple(map(tuple, polygon))
    # make the mask
    img = Image.new('L', (width, height), 0)
    ImageDraw.Draw(img).polygon(vertices_tuple, outline=1, fill=1)

    return np.array(img, dtype='int32')


def _get_element_mask(elinfo, slide_annotations):
    """Get coordinates and mask for annotation element.

    (Internal)

    """
    element = slide_annotations[int(elinfo['annidx'])][
        'annotation']['elements'][int(elinfo['elementidx'])]

    if elinfo['type'] == 'polyline':
        coords = np.array(element['points'])[:, :-1]
    elif elinfo['type'] == 'rectangle':
        infoDict = get_rotated_rectangular_coords(
            roi_center=element['center'], roi_width=element['width'],
            roi_height=element['height'], roi_rotation=element['rotation'])
        coords = infoDict['roi_corners']
    else:
        msg = 'cannot create mask from point annotation!'
        raise Exception(msg)

    mask = create_mask_from_coords(coords)
    return coords, mask


def _add_element_to_roi(elinfo, ROI, GT_code, element_mask, roiinfo):
    """Add single polygon to ROI given mask (Internal)."""
    ymin = int(elinfo.ymin - roiinfo['YMIN'])
    ymax = int(elinfo.ymax - roiinfo['YMIN'])
    xmin = int(elinfo.xmin - roiinfo['XMIN'])
    xmax = int(elinfo.xmax - roiinfo['XMIN'])
    patch = ROI[ymin:ymax, xmin:xmax]

    # to account for non-integer pixels
    if element_mask.shape != patch.shape:
        element_mask = np.pad(element_mask, pad_width=(
            (patch.shape[0] - element_mask.shape[0], 0),
            (patch.shape[1] - element_mask.shape[1], 0)), mode='constant')

    # add to ROI mask
    patch[element_mask > 0] = GT_code
    ROI[ymin:ymax, xmin:xmax] = patch.copy()

    element = {
        'mask': element_mask,
        'xmin': xmin, 'xmax': xmax,
        'ymin': ymin, 'ymax': ymax,
    }

    return ROI, element


def _get_and_add_element_to_roi(
        elinfo, slide_annotations, ROI, roiinfo, roi_polygon, GT_code,
        use_shapely=True, verbose=True, monitorPrefix=''):
    """Get element coords and mask and add to ROI (Internal)."""
    try:
        coords, element_mask = _get_element_mask(
            elinfo=elinfo, slide_annotations=slide_annotations)

        ADD_TO_ROI = True

        # ignore if outside ROI (precise)
        if use_shapely:
            el_polygon = Polygon(coords)
            if el_polygon.distance(roi_polygon) > 2:
                if verbose:
                    print('%s: OUTSIDE ROI.' % monitorPrefix)
                ADD_TO_ROI = False

        # Add element to ROI mask
        if ADD_TO_ROI:
            ROI, _ = _add_element_to_roi(
                elinfo=elinfo, ROI=ROI, GT_code=GT_code,
                element_mask=element_mask, roiinfo=roiinfo)

    except Exception as e:
        if verbose:
            print('%s: ERROR! (see below)' % monitorPrefix)
            print(e)
    return ROI


def delete_annotations_in_slide(gc, slide_id):
    """Delete all annotations in a slide."""
    existing_annotations = gc.get('/annotation/item/' + slide_id)
    for ann in existing_annotations:
        gc.delete('/annotation/%s' % ann['_id'])


def _simple_add_element_to_roi(
        elinfo, ROI, roiinfo, GT_code, element=None,
        verbose=True, monitorPrefix=''):
    """Get element coords and mask and add to ROI (Internal)."""

    def _process_coords(k):
        return np.array([int(j) for j in elinfo[k].split(',')])[..., None]

    try:
        if element is None:
            coords = np.concatenate([
                _process_coords(k) for k in ('coords_x', 'coords_y')], 1)
            element_mask = create_mask_from_coords(coords)
        else:
            element_mask = element['mask']

        # Add element to ROI mask
        ROI, element = _add_element_to_roi(
            elinfo=elinfo, ROI=ROI, GT_code=GT_code,
            element_mask=element_mask, roiinfo=roiinfo)

    except Exception as e:
        if verbose:
            print('%s: ERROR! (see below)' % monitorPrefix)
            print(e)
    return ROI, element



import numpy as np


def embed_boundaries(im_input, im_perim, color=None):
    """Embeds object boundaries into an RGB color, grayscale or binary
    image, returning a color rendering of the image and object
    boundaries.

    Takes as input a grayscale or color image, a perimeter mask of object
    boundaries, and an RGB triplet, and embeds the object boundaries into
    the input image at the prescribed color. Returns a color RGB image of
    type unsigned char. If the input image is type double, and has pixels
    inside range [0, 1], then it will be scaled to the range [0, 255].
    Otherwise it will be assumed to be in the range of an unsigned char
    image.

    Parameters
    ----------
    im_input : array_like
        A color or grayscale image.
    im_perim : array_like
        A binary image where object perimeter pixels have value 1, and
        non-perimeter pixels have value 0.
    color : array_like
        A 1 x 3 array of RGB values in the range [0, 255].

    Returns
    -------
    im_embed : array_like
        A color image of type unsigned char where boundary pixels take
        on the color defined by the RGB-triplet 'color'.

    See Also
    --------
    histomicstk.segmentation.label.perimeter

    """
    color = [255, 0, 0] if color is None else color

    # check for consistent shapes between 'im_input' and 'im_perim'
    if im_input.shape[0:2] != im_perim.shape[0:2]:
        msg = "'im_input' and 'im_perim' must have same shape"
        raise ValueError(msg)

    # determine image type
    if np.issubclass_(im_input.dtype.type, np.float_):
        if im_input.max() < 1.0:
            im_input *= 255
            im_input = im_input.astype(np.uint8)
    elif np.issubclass_(im_input.dtype.type, np.bool_):
        im_input = im_input.astype(np.uint8)
        im_input *= 255

    # determine if image is grayscale or RGB
    if len(im_input.shape) == 3:  # color image
        Red = im_input[:, :, 0].copy()
        Green = im_input[:, :, 1].copy()
        Blue = im_input[:, :, 2].copy()
    elif len(im_input.shape) == 2:  # grayscale image
        Red = im_input.copy()
        Green = im_input.copy()
        Blue = im_input.copy()

    # embed boundaries
    Red[im_perim > 0] = color[0]
    Green[im_perim > 0] = color[1]
    Blue[im_perim > 0] = color[2]

    # generate output image
    im_embed = np.dstack((Red[:, :, np.newaxis],
                          Green[:, :, np.newaxis],
                          Blue[:, :, np.newaxis])).astype(np.uint8)

    # concatenate channels to form output
    return im_embed



from histomicstk.cli import utils as cli_utils
from histomicstk.segmentation import label as htk_label

from .compute_fsd_features import compute_fsd_features
from .compute_gradient_features import compute_gradient_features
from .compute_haralick_features import compute_haralick_features
from .compute_intensity_features import compute_intensity_features
from .compute_morphometry_features import compute_morphometry_features


def compute_nuclei_features(im_label, im_nuclei=None, im_cytoplasm=None,
                            fsd_bnd_pts=128, fsd_freq_bins=6, cyto_width=8,
                            num_glcm_levels=32,
                            morphometry_features_flag=True,
                            fsd_features_flag=True,
                            intensity_features_flag=True,
                            gradient_features_flag=True,
                            haralick_features_flag=True,
                            tile_info=None,
                            im_nuclei_seg_mask=None,
                            format=None,
                            return_nuclei_annotation=False,
                            ):
    """
    Calculates features for nuclei classification

    Parameters
    ----------
    im_label : array_like
        A labeled mask image wherein intensity of a pixel is the ID of the
        object it belongs to. Non-zero values are considered to be foreground
        objects.

    im_nuclei : array_like
        Nucleus channel intensity image.

    im_cytoplasm : array_like
        Cytoplasm channel intensity image.

    fsd_bnd_pts : int, optional
        Number of points for boundary resampling to calculate fourier
        descriptors. Default value = 128.

    fsd_freq_bins : int, optional
        Number of frequency bins for calculating FSDs. Default value = 6.

    cyto_width : float, optional
        Estimated width of the ring-like neighborhood region around each
        nucleus to be considered as its cytoplasm. Default value = 8.

    num_glcm_levels: int, optional
        An integer specifying the number of gray levels For example, if
        `NumLevels` is 32,  the intensity values of the input image are
        scaled so they are integers between 0 and 31.  The number of gray
        levels determines the size of the gray-level co-occurrence matrix.

        Default: 32

    morphometry_features_flag : bool, optional
        A flag that can be used to specify whether or not to compute
        morphometry (size and shape) features.
        See histomicstk.features.compute_morphometry_features for more details.

    fsd_features_flag : bool, optional
        A flag that can be used to specify whether or not to compute
        Fouried shape descriptor (FSD) features.
        See `histomicstk.features.compute_fsd_features` for more details.

    intensity_features_flag : bool, optional
        A flag that can be used to specify whether or not to compute
        intensity features from the nucleus and cytoplasm channels.
        See `histomicstk.features.compute_fsd_features` for more details.

    gradient_features_flag : bool, optional
        A flag that can be used to specify whether or not to compute
        gradient/edge features from intensity and cytoplasm channels.
        See `histomicstk.features.compute_gradient_features` for more details.

    haralick_features_flag : bool, optional
        A flag that can be used to specify whether or not to compute
        haralick features from intensity and cytoplasm channels.
        See `histomicstk.features.compute_haralick_features` for more details.

    return_nuclei_annotation :  bool, optional
        Returns the nuclei annotation if kept True

    Returns
    -------
    fdata : pandas.DataFrame
        A pandas data frame containing the features listed below for each
        object/label
    nuclei_annot_list : List
        List containing the boundaries of segmented nuclei in the input image.

    Notes
    -----
    List of features computed by this function

    Identifier
        Location of the nucleus and its code in the input labeled mask.
        Columns are prefixed by *Identifier.*. These include ...

        Identifier.Label (int) - nucleus label in the input labeled mask

        Identifier.Xmin (int) - Left bound

        Identifier.Ymin (int) - Upper bound

        Identifier.Xmax (int) - Right bound

        Identifier.Ymax (int) - Lower bound

        Identifier.CentroidX (float) - X centroid (columns)

        Identifier.CentroidY (float) - Y centroid (rows)

        Identifier.WeightedCentroidX (float) - intensity-weighted X centroid

        Identifier.WeightedCentroidY (float) - intensity-weighted Y centroid

    Morphometry (size, shape, and orientation) features of the nuclei
        See histomicstk.features.compute_morphometry_features for more details.
        Feature names prefixed by *Size.*, *Shape.*, or *Orientation.*.

    Fourier shape descriptor features
        See `histomicstk.features.compute_fsd_features` for more details.
        Feature names are prefixed by *FSD*.

    Intensity features for the nucleus and cytoplasm channels
        See `histomicstk.features.compute_fsd_features` for more details.
        Feature names are prefixed by *Nucleus.Intensity.* for nucleus features
        and *Cytoplasm.Intensity.* for cytoplasm features.

    Gradient/edge features for the nucleus and cytoplasm channels
        See `histomicstk.features.compute_gradient_features` for more details.
        Feature names are prefixed by *Nucleus.Gradient.* for nucleus features
        and *Cytoplasm.Gradient.* for cytoplasm features.

    Haralick features for the nucleus and cytoplasm channels
        See `histomicstk.features.compute_haralick_features` for more details.
        Feature names are prefixed by *Nucleus.Haralick.* for nucleus features
        and *Cytoplasm.Haralick.* for cytoplasm features.

    See Also
    --------
    histomicstk.features.compute_morphometry_features,
    histomicstk.features.compute_fsd_features,
    histomicstk.features.compute_intensity_features,
    histomicstk.features.compute_gradient_features,
    histomicstk.features.compute_haralick_features

    """
    import pandas as pd
    from skimage.measure import regionprops

    # sanity checks
    if any([
        intensity_features_flag,
        gradient_features_flag,
        haralick_features_flag,
    ]):
        assert im_nuclei is not None, 'You must provide nuclei intensity!'

    # TODO: this pipeline uses loops a lot. For each set of features it
    #  iterates over all nuclei, which may become an issue when one needs to
    #  do this for lots and lots of slides and 10^6+ nuclei. Consider
    #  improving efficiency in the future somehow (cython? reuse? etc)

    feature_list = []

    # get the objects in im_label
    nuclei_props = regionprops(im_label, intensity_image=im_nuclei)

    im_nuclei_bool = im_nuclei is not None

    data = []

    def process_nucleus(nprop, im_nuclei_bool):
        for nprop in nuclei_props:
            row = {
                'Label': nprop.label,
                'Identifier.Xmin': nprop.bbox[1],
                'Identifier.Ymin': nprop.bbox[0],
                'Identifier.Xmax': nprop.bbox[3],
                'Identifier.Ymax': nprop.bbox[2],
                'Identifier.CentroidX': nprop.centroid[1],
                'Identifier.CentroidY': nprop.centroid[0],
            }
            if im_nuclei_bool:
                wcy, wcx = nprop.weighted_centroid
                row['Identifier.WeightedCentroidX'] = wcx
                row['Identifier.WeightedCentroidY'] = wcy
            data.append(row)
        return pd.DataFrame(data)

    # create the DataFrame in one step
    idata = process_nucleus(nuclei_props, im_nuclei_bool)
    feature_list.append(idata)

    def conditional(flag, func, args, kwargs, prefix=None):
        if flag:
            output = func(*args, **kwargs)
            if prefix:
                output.columns = [prefix + col for col in output.columns]
            return output
        else:
            return pd.DataFrame()  # return an empty DataFrame if condition is not met

    # compute cytoplasm mask
    if im_cytoplasm is not None:

        cyto_mask = htk_label.dilate_xor(im_label, neigh_width=cyto_width)

        cyto_props = regionprops(cyto_mask, intensity_image=im_cytoplasm)

        # ensure that cytoplasm props order corresponds to the nuclei
        lablocs = {v['label']: i for i, v in enumerate(cyto_props)}
        cyto_props = [cyto_props[lablocs[v['label']]] if v['label'] in lablocs else None
                      for v in nuclei_props]

    # compute morphometry features
    feature_list.append(conditional(
        morphometry_features_flag,
        compute_morphometry_features,
        [im_label], {'rprops': nuclei_props},
    ))

    # compute FSD features
    feature_list.append(conditional(
        fsd_features_flag,
        compute_fsd_features,
        [im_label, fsd_bnd_pts, fsd_freq_bins, cyto_width], {'rprops': nuclei_props},
    ))

    # compute nuclei intensity features
    feature_list.append(conditional(
        intensity_features_flag,
        compute_intensity_features,
        [im_label, im_nuclei], {'rprops': nuclei_props}, prefix='Nucleus.',
    ))

    # compute cytoplasm intensity features
    if im_cytoplasm is not None:
        feature_list.append(conditional(
            intensity_features_flag,
            compute_intensity_features,
            [cyto_mask, im_cytoplasm], {'rprops': cyto_props}, prefix='Cytoplasm.',
        ))

    # compute nuclei gradient features
    feature_list.append(conditional(
        gradient_features_flag,
        compute_gradient_features,
        [im_label, im_nuclei], {'rprops': nuclei_props}, prefix='Nucleus.',
    ))

    # compute cytoplasm gradient features
    if im_cytoplasm is not None:
        feature_list.append(conditional(
            gradient_features_flag,
            compute_gradient_features,
            [cyto_mask, im_cytoplasm], {'rprops': cyto_props}, prefix='Cytoplasm.',
        ))

    # compute nuclei haralick features
    feature_list.append(conditional(
        haralick_features_flag,
        compute_haralick_features,
        [im_label, im_nuclei], {'num_levels': num_glcm_levels, 'rprops': nuclei_props},
        prefix='Nucleus.',
    ))

    # compute cytoplasm haralick features
    if im_cytoplasm is not None:
        feature_list.append(conditional(
            haralick_features_flag,
            compute_haralick_features,
            [cyto_mask, im_cytoplasm], {'num_levels': num_glcm_levels, 'rprops': cyto_props},
            prefix='Cytoplasm.',
        ))

    # Merge all features
    fdata = pd.concat(feature_list, axis=1)

    if return_nuclei_annotation:
        # Create nuclei segmentation with the generated regionprops
        nuclei_annot_list, selected_rows = cli_utils.create_tile_nuclei_annotations(
            im_nuclei_seg_mask, tile_info, format, nuclei_props)

        # Drop all rows which are not found in nuclei detection
        fdata = fdata[fdata.index.isin(selected_rows)]
        return fdata, nuclei_annot_list

    return fdata



"""
Created on Tue Sep 24 00:43:04 2019.

@author: mtageld
"""
import datetime
import logging
import os


class Print_and_log:
    """Print to screen and/or log if conditions are satisfied (Internal)."""

    def __init__(self, verbose=True, logger=None):
        """Init. This is for PEP compliance."""
        self.verbose = verbose
        self.keep_log = logger is not None
        if self.keep_log:
            self.logger = logger

    def _print(self, text):
        if self.verbose:
            print(text)
        if self.keep_log:
            self.logger.info(text)


class Base_HTK_Class:
    """Just a base class with preferred behavior to inherit."""

    def __init__(self, default_attr=None, more_allowed_attr=None, **kwargs):
        """Init base HTK class."""
        default_attr = {} if default_attr is None else default_attr
        more_allowed_attr = [] if more_allowed_attr is None else more_allowed_attr

        da = {
            'verbose': 1,
            'monitorPrefix': '',
            'logging_savepath': None,
            'suppress_warnings': False,
        }
        default_attr.update(
            (k, v) for k, v in da.items() if k not in default_attr.keys())

        # See: https://stackoverflow.com/questions/8187082/how-can-you-set-...
        # class-attributes-from-variable-arguments-kwargs-in-python
        allowed_attr = list(default_attr.keys()) + more_allowed_attr
        default_attr.update(kwargs)
        self.__dict__.update(
            (k, v) for k, v in default_attr.items() if k in allowed_attr)

        # To NOT silently ignore rejected keys
        rejected_keys = set(kwargs.keys()) - set(allowed_attr)
        if rejected_keys:
            msg = f'Invalid arguments in constructor:{rejected_keys}'
            raise ValueError(
                msg)

        # configure logger
        self.keep_log = self.logging_savepath is not None
        if self.keep_log:
            logger = logging.getLogger()
            self.logname = os.path.join(
                self.logging_savepath,
                datetime.datetime.now().strftime('%Y-%m-%d_%H-%M') + '.log')
            logging.basicConfig(filename=self.logname, level=logging.INFO)
        else:
            logger = None

        # verbosity control
        self.cpr1 = Print_and_log(verbose=self.verbose >= 1, logger=logger)
        self._print1 = self.cpr1._print
        self.cpr2 = Print_and_log(verbose=self.verbose >= 2, logger=logger)
        self._print2 = self.cpr2._print
        if self.keep_log:
            self._print1('Saving logs to: %s' % self.logname)



import os
import sys

import pytest

from histomicstk.utils.girder_convenience_utils import (
    get_absolute_girder_folderpath, revert_annotations_in_slide,
    update_permissions_for_annotations_in_slide,
    update_styles_for_annotations_in_slide)

thisDir = os.path.dirname(os.path.realpath(__file__))
sys.path.insert(0, os.path.join(thisDir, '../../../'))
from tests.htk_test_utilities import girderClient  # noqa


class Cfg:
    def __init__(self):
        self.gc = None
        self.iteminfo = None
        self.posted_folder = None


cfg = Cfg()


class TestGirderConvenience:
    """Test utilities for interaction with girder."""

    # pytest runs tests in the order they appear in the module
    @pytest.mark.usefixtures('girderClient')  # noqa
    def test_prep(self, girderClient):  # noqa

        cfg.gc = girderClient

        # get original item
        original_iteminfo = cfg.gc.get('/item', parameters={'text': 'TCGA-A2-A0YE-01Z-00-DX1'})[0]

        # create a sample folder
        cfg.posted_folder = cfg.gc.post(
            '/folder', data={
                'parentId': original_iteminfo['folderId'],
                'name': 'test-convenience',
            })

        # copy the item so that everything we do here does not affect
        # other modules that use that item
        cfg.iteminfo = cfg.gc.post(
            '/item/%s/copy' % original_iteminfo['_id'], data={
                'name': 'test_slide_gcutils',
                'copyAnnotations': True,
            })

    def test_get_absolute_girder_path(self):
        # now get and check absolute path
        fpath = get_absolute_girder_folderpath(
            gc=cfg.gc, folder_info=cfg.posted_folder)
        assert fpath == 'Public/test-convenience/'

    def test_update_permissions_for_annotations_in_slide(self):
        admininfo = cfg.gc.get('/user', parameters={
            'text': 'admin'})[0]

        # params to pass to update_permissions_for_annotation()
        update_params = {
            'users_to_add': [
                {'level': 2, 'id': admininfo['_id'],
                 'login': admininfo['login']},
            ],
            'groups_to_add': [],
            'replace_original_groups': True,
            'replace_original_users': True,
        }
        resps = update_permissions_for_annotations_in_slide(
            gc=cfg.gc, slide_id=cfg.iteminfo['_id'],
            # monitorPrefix='test_update_permissions_for_annotations_in_slide',
            **update_params,
        )
        assert len(resps) >= 8
        assert resps[0]['access']['users'] == [
            {'flags': [], 'id': admininfo['_id'], 'level': 2},
        ]
        assert resps[0]['access']['groups'] == []

    def test_update_styles_for_annotations_in_slide(self):
        resps = update_styles_for_annotations_in_slide(
            gc=cfg.gc, slide_id=cfg.iteminfo['_id'],
            changes={
                'roi': {
                    'group': 'modified_roi',
                    'lineColor': 'rgb(0,0,255)',
                    'fillColor': 'rgba(0,0,255,0.3)',
                },
            },
            # monitorPrefix='test_update_styles_for_annotations_in_slide',
        )
        modified = [j for j in resps if j is not None]
        assert len(modified) == 3
        assert 'modified_roi' in modified[0]['groups']

    def test_revert_annotations_in_slide(self):
        # delete elements for one annotation
        anns = cfg.gc.get('/annotation/item/%s' % cfg.iteminfo['_id'])
        anns[0]['annotation']['elements'] = []
        _ = cfg.gc.put(
            '/annotation/%s' % anns[0]['_id'], json=anns[0]['annotation'])

        # now revert
        resps = revert_annotations_in_slide(
            gc=cfg.gc, slide_id=cfg.iteminfo['_id'],
            revert_to_nonempty_elements=False,
            only_revert_if_empty=False,
            # monitorPrefix='test_revert_annotations_in_slide',
        )
        modified = [j for j in resps if len(j) > 0]
        assert len(modified) >= 1
        ann = cfg.gc.get('/annotation/%s' % modified[0]['_id'])
        assert len(ann['annotation']['elements']) >= 1



"""Functions to pre-process histopathology images."""
# import sub-packages to support nested calls
from . import (augmentation, color_conversion, color_deconvolution,
               color_normalization)

# list out things that are available for public use
__all__ = (
    'color_conversion',
    'color_deconvolution',
    'color_normalization',
    'augmentation',
)



import numpy as np

from .condense import condense


def width_open(im_label, width):
    """Removes thin objects from label image using maximum of distance
    transform values within each object.

    Parameters
    ----------
    im_label : array_like
        A uint32 type label image generated by segmentation methods.
    width : int
        width threshold for objects. Objects with fewer than 'Area' pixels will
        be zeroed to merge with background.

    Notes
    -----
    Objects are assumed to have positive nonzero values. A binary mask is
    generated for each object setting all other objects to the background value
    (0). The maximum chamfered distance transform value of this mask is used
    to represent the object width.

    Returns
    -------
    im_thinned : array_like
        A uint32 label where objects with pixels < Area are removed.

    See Also
    --------
    histomicstk.segmentation.label.condense,
    histomicstk.segmentation.label.shuffle,
    histomicstk.segmentation.label.split,
    histomicstk.segmentation.label.area_open

    """
    import scipy.ndimage as ndi

    # copy input image
    im_thinned = im_label.copy()

    # condense label image
    if np.unique(im_thinned).size - 1 != im_thinned.max():
        im_thinned = condense(im_thinned)

    # get locations of objects in initial label image
    Locations = ndi.find_objects(im_thinned)

    # iterate through objects, calculating distances where needed
    for i in np.arange(1, len(Locations) + 1):

        # extract object from label image
        W = im_thinned[Locations[i - 1]]

        # embed into mask with boundary
        Mask = np.zeros((W.shape[0] + 2, W.shape[1] + 2), dtype=bool)
        Mask[1:-1, 1:-1] = W == i

        # calculate distance transform of mask
        D = ndi.distance_transform_cdt(Mask, metric='taxicab')

        # get max distance
        Max = D.max()

        # zero label mask of object 'i'
        if Max < width:
            W[W == i] = 0

    # condense to fill gaps
    im_thinned = condense(im_thinned)

    return im_thinned



"""
Created on Thu Dec 12 13:19:18 2019

@author: tageldim
"""
import copy
import json
import os

from sqlalchemy import create_engine
from sqlalchemy.types import Boolean, Integer, String

from histomicstk.annotations_and_masks.annotation_and_mask_utils import \
    parse_slide_annotations_into_tables
from histomicstk.utils.girder_convenience_utils import \
    get_absolute_girder_folderpath
from histomicstk.workflows.workflow_runner import (Slide_iterator,
                                                   Workflow_runner)

# Helper functions


def _add_item_to_sqlite(dbcon, item):
    from pandas import DataFrame

    # modify item info to prep for appending to sqlite table
    item_info = copy.deepcopy(item)
    item_info['largeImage'] = str(item_info['largeImage'])

    item_info_dtypes = {
        '_id': String(),
        '_modelType': String(),
        'baseParentId': String(),
        'baseParentType': String(),
        'copyOfItem': String(),
        'created': String(),
        'creatorId': String(),
        'description': String(),
        'folderId': String(),
        'largeImage': String(),
        'name': String(),
        'size': Integer(),
        'updated': String(),
    }

    # in case anything is not in the schema, drop it
    item_info = {
        k: v for k, v in item_info.items()
        if k in item_info_dtypes.keys()}

    # convert to df and add to items table
    item_info_df = DataFrame.from_dict(item_info, orient='index').T
    item_info_df.to_sql(
        name='items', con=dbcon, if_exists='append',
        dtype=item_info_dtypes, index=False)


def _add_folder_to_sqlite(dbcon, folder_info):
    from pandas import DataFrame

    # modify folder info to prep for appending to sqlite table
    folder_info_dtypes = {
        '_accessLevel': Integer(),
        '_id': String(),
        '_modelType': String(),
        'baseParentId': String(),
        'baseParentType': String(),
        'created': String(),
        'creatorId': String(),
        'description': String(),
        'name': String(),
        'parentCollection': String(),
        'parentId': String(),
        'public': Boolean(),
        'size': Integer(),
        'updated': String(),
        'folder_path': String(),
    }

    # in case anything is not in the schema, drop it
    folder_info = {
        k: v for k, v in folder_info.items()
        if k in folder_info_dtypes.keys()}

    # convert to df and add to items table
    folder_info_df = DataFrame.from_dict(folder_info, orient='index').T
    folder_info_df.to_sql(
        name='folders', con=dbcon, if_exists='append',
        dtype=folder_info_dtypes, index=False)


def _add_annotation_docs_to_sqlite(dbcon, annotation_docs, item):
    # add full item path for convenience
    annotation_docs.loc[:, 'item_name'] = item['name']

    # save tables to sqlite
    annotation_docs.to_sql(
        name='annotation_docs', con=dbcon, if_exists='append',
        dtype={
            'annotation_girder_id': String(),
            '_modelType': String(),
            '_version': Integer(),
            'itemId': String(),
            'item_name': String(),
            'created': String(),
            'creatorId': String(),
            'public': Boolean(),
            'updated': String(),
            'updatedId': String(),
            'groups': String(),
            'element_count': Integer(),
            'element_details': Integer()},
        index=False,
    )


def _add_annotation_elements_to_sqlite(dbcon, annotation_elements):
    # drop index relative to JSON since its pretty arbitrary and would
    # change if the same girder client was used to get annotations twice
    # the actual girder ID string is what really matters and should be used
    annotation_elements.drop(
        labels=['annidx', 'elementidx'], axis=1, inplace=True)

    annotation_elements.to_sql(
        name='annotation_elements', con=dbcon, if_exists='append',
        dtype={
            'annotation_girder_id': String(),
            'element_girder_id': String(),
            'type': String(),
            'group': String(),
            'label': String(),
            'color': String(),
            'xmin': Integer(),
            'xmax': Integer(),
            'ymin': Integer(),
            'ymax': Integer(),
            'bbox_area': Integer(),
            'coords_x': String(),
            'coords_y': String()},
        index=False,
    )


def parse_annotations_to_local_tables(
        item, annotations, local, monitorPrefix='',
        save_csv=True, save_sqlite=False, dbcon=None):
    """Parse loaded annotations for slide into tables.

    Parameters
    ----------
    item : dict
        girder response with item information

    annotations : dict
        loaded annotations

    local : str
        local directory

    save_csv : bool
        whether to use histomicstk.annotations_and_masks.annotation_and_mask.
        parse_slide_annotations_into_tables() to get a tabular representation
        (including some simple calculations like bounding box) and save
        the output as two csv files, one representing the annotation documents
        and the other representing the actual annotation elements (polygons).

    save_sqlite : bool
        whether to save the backup into an sqlite database

    dbcon : sqlalchemy.create_engine.connect() object
        IGNORE THIS PARAMETER!! This is used internally.

    monitorPrefix : str
        text to prepend to printed statements

    """
    print('%s: parse to tables' % monitorPrefix)
    savepath_base = os.path.join(local, item['name'])
    annotation_docs, annotation_elements = \
        parse_slide_annotations_into_tables(annotations)

    if save_csv:
        annotation_docs.to_csv(savepath_base + '_docs.csv')
        annotation_elements.to_csv(savepath_base + '_elements.csv')

    if save_sqlite:
        assert dbcon is not None, 'You must connect to database first!'
        _add_annotation_docs_to_sqlite(dbcon, annotation_docs, item)
        _add_annotation_elements_to_sqlite(dbcon, annotation_elements)


# Workflow at a single slide level


def dump_annotations_workflow(
        gc, slide_id, local, monitorPrefix='',
        save_json=True, save_sqlite=False, dbcon=None,
        callback=None, callback_kwargs=None):
    """Dump annotations for single slide into the local folder.

    Parameters
    ----------
    gc : girder_client.GirderClient
        authenticated girder client instance

    slide_id : str
        girder id of item (slide)

    monitorPrefix : str
        prefix to monitor string

    local : str
        local path to dump annotations

    save_json : bool
        whether to dump annotations as json file

    save_sqlite : bool
        whether to save the backup into an sqlite database

    dbcon : sqlalchemy.create_engine.connect() object
        IGNORE THIS PARAMETER!! This is used internally.

    callback : function
        function to call that takes in AT LEAST the following params
        - item: girder response with item information
        - annotations: loaded annotations
        - local: local directory
        - monitorPrefix: string

    callback_kwargs : dict
        kwargs to pass along to callback

    """
    callback_kwargs = callback_kwargs or {}
    try:
        item = gc.get('/item/%s' % slide_id)

        savepath_base = os.path.join(local, item['name'])

        # dump item information json
        if save_json:
            print('%s: save item info' % monitorPrefix)
            with open(savepath_base + '.json', 'w') as fout:
                json.dump(item, fout)

        # save folder info to sqlite
        if save_sqlite:
            _add_item_to_sqlite(dbcon, item)

        # pull annotation
        print('%s: load annotations' % monitorPrefix)
        annotations = gc.get('/annotation/item/' + item['_id'])

        if annotations is not None:

            # dump annotations to JSON in local folder
            if save_json:
                print('%s: save annotations' % monitorPrefix)
                with open(savepath_base + '_annotations.json', 'w') as fout:
                    json.dump(annotations, fout)

            # run callback
            if callback is not None:
                print('%s: run callback' % monitorPrefix)
                callback(
                    item=item, annotations=annotations, local=local,
                    dbcon=dbcon, monitorPrefix=monitorPrefix,
                    **callback_kwargs)

    except Exception as e:
        print(str(e))


# Main method


def dump_annotations_locally(
        gc, folderid, local, save_json=True,
        save_sqlite=False, dbcon=None,
        callback=None, callback_kwargs=None):
    """Dump annotations of folder and subfolders locally recursively.

    This reproduces this tiered structure locally and (possibly) dumps
    annotations there. Adapted from Lee A.D. Cooper

    Parameters
    ----------
    gc : girder_client.GirderClient
        authenticated girder client instance

    folderid : str
        girder id of source (base) folder

    local : str
        local path to dump annotations

    save_json : bool
        whether to dump annotations as json file

    save_sqlite : bool
        whether to save the backup into an sqlite database

    dbcon : sqlalchemy.create_engine.connect() object
        IGNORE THIS PARAMETER!! This is used internally.

    callback : function
        function to call that CAN accept AT LEAST the following params
        - item: girder response with item information
        - annotations: loaded annotations
        - local: local directory
        - monitorPrefix: string
        - dbcon: sqlalchemy.create_engine.connect() object
        You can just add kwargs at the end of your callback definition
        for simplicity.

    callback_kwargs : dict
        kwargs to pass along to callback. DO NOT pass any of the parameters
        item, annotations, local, monitorPrefix, or dbcon as these will be
        internally passed. Just include any specific parameters for the
        callback. See parse_annotations_to_local_tables() above for
        an example of a callback and the unir test of this function.

    """
    callback_kwargs = callback_kwargs or {}
    assert save_json or save_sqlite, 'must save results somehow!'
    monitor = os.path.basename(local)

    # get folder info
    folder_info = gc.get('folder/%s' % folderid)
    folder_info['folder_path'] = get_absolute_girder_folderpath(
        gc=gc, folder_info=folder_info)

    # connect to sqlite database -- only first stack does this
    if save_sqlite and (dbcon is None):
        db_path = os.path.join(local, folder_info['name'] + '.sqlite')
        sql_engine = create_engine('sqlite:///' + db_path, echo=False)
        dbcon = sql_engine.connect()

    # save folder information json
    if save_json:
        print('%s: save folder info' % monitor)
        savepath = os.path.join(local, folder_info['name'] + '.json')
        with open(savepath, 'w') as fout:
            json.dump(folder_info, fout)

    # save folder info to sqlite
    if save_sqlite:
        _add_folder_to_sqlite(dbcon, folder_info)

    # pull annotations for each slide in folder
    workflow_runner = Workflow_runner(
        slide_iterator=Slide_iterator(
            gc, source_folder_id=folderid,
            keep_slides=None,
        ),
        workflow=dump_annotations_workflow,
        workflow_kwargs={
            'gc': gc,
            'local': local,
            'save_json': save_json,
            'save_sqlite': save_sqlite,
            'dbcon': dbcon,
            'callback': callback,
            'callback_kwargs': callback_kwargs,
        },
        monitorPrefix=monitor)

    workflow_runner.run()

    # for each subfolder, create a new folder locally and call self
    for folder in gc.listFolder(parentId=folderid):

        # create folder in local
        new_folder = os.path.join(local, folder['name'])
        os.mkdir(new_folder)

        # call self with same parameters
        dump_annotations_locally(
            gc=gc, folderid=folder['_id'], local=new_folder,
            save_json=save_json, save_sqlite=save_sqlite, dbcon=dbcon,
            callback=callback, callback_kwargs=callback_kwargs)



import numpy as np


def rag_add_layer(adj_mat):
    """Adds an additional layer of dependence to a region adjacency graph,
    connecting each node to the neighbors of its immediate neighbors.

    Parameters
    ----------
    adj_mat : array_like
        A binary matrix of size N x N, where N is the number of objects in
        Label. A value of 'True' at adj_mat(i,j) indicates that objects 'i'
        and 'j' are neighbors.

    Returns
    -------
    Layered : array_like
        A version of 'adj_mat' with additional edges to connect 2-neighbors.

    See Also
    --------
    histomicstk.segmentation.rag

    """
    # initialize output
    Layered = adj_mat.copy()

    # iterate through each object, adding edges to neighbors of neighbors
    for i in range(adj_mat.shape[0]):

        # get immediate neighbors
        Neighbors = np.nonzero(adj_mat[i, ])[0].flatten()

        # for each immediate neighbor, add second neighbors
        for j in range(Neighbors.size):
            Hops = np.nonzero(adj_mat[Neighbors[j], ])[0].flatten()
            Layered[Hops, i] = True
            Layered[i, Hops] = True

    # return result
    return Layered



"""
Created on Fri Oct 18 02:31:32 2019.

@author: mtageld
"""
import numpy as np

from histomicstk.preprocessing.color_deconvolution import \
    complement_stain_matrix
from histomicstk.preprocessing.color_deconvolution.color_convolution import \
    color_convolution
from histomicstk.preprocessing.color_deconvolution.color_deconvolution import (
    color_deconvolution_routine, stain_unmixing_routine)
from histomicstk.preprocessing.color_deconvolution.stain_color_map import \
    stain_color_map


def deconvolution_based_normalization(
        im_src, W_source=None, W_target=None, im_target=None,
        stains=None, mask_out=None, stain_unmixing_routine_params=None):
    """Perform color normalization using color deconvolution to transform the.

    ... color characteristics of an image to a desired standard.
    After the image is deconvolved into its component stains (eg, H&E), it is
    convolved with a stain column vectors matrix from the target image from
    which the color characteristics need to be transferred.

    Parameters
    ----------
    im_src : array_like
        An RGB image (m x n x 3) to color normalize

    W_source : np array, default is None
        A 3x3 matrix of source stain column vectors. Only provide this
        if you know the stains matrix in advance (unlikely) and would
        like to perform supervised deconvolution. If this is not provided,
        stain_unmixing_routine() is used to estimate W_source.

    W_target : np array, default is None
        A 3x3 matrix of target stain column vectors. If not provided,
        and im_target is also not provided, the default behavior is to use
        histomicstk.preprocessing.color_deconvolution.stain_color_map
        to provide an idealized target matrix.

    im_target : array_like, default is None
        An RGB image (m x n x 3) that has good color properties that ought to
        be transferred to im_src. If you provide this parameter, im_target
        will be used to extract W_target and the W_target parameter will
        be ignored.

    stains : list, optional
        List of stain names (order is important). Default is H&E. This is
        particularly relevant in macenco where the order of stains is not
        preserved during stain unmixing, so this method uses
        histomicstk.preprocessing.color_deconvolution.find_stain_index
        to reorder the stains matrix to the order provided by this parameter

    mask_out : array_like, default is None
        if not None, should be (m x n) boolean numpy array.
        This parameter ensures exclusion of non-masked areas from calculations
        and normalization. This is relevant because elements like blood,
        sharpie marker, white space, etc may throw off the normalization.

    stain_unmixing_routine_params : dict, default is empty dict
        k,v for stain_unmixing_routine().

    Returns
    -------
    array_like
        Color Normalized RGB image (m x n x 3)


    See Also
    --------
    histomicstk.preprocessing.color_deconvolution.color_deconvolution_routine
    histomicstk.preprocessing.color_convolution.color_convolution

    References
    ----------
    .. [#] Van Eycke, Y. R., Allard, J., Salmon, I., Debeir, O., &
           Decaestecker, C. (2017).  Image processing in digital pathology: an
           opportunity to solve inter-batch variability of immunohistochemical
           staining.  Scientific Reports, 7.
    .. [#] Macenko, M., Niethammer, M., Marron, J. S., Borland, D.,
           Woosley, J. T., Guan, X., ... & Thomas, N. E. (2009, June).
           A method for normalizing histology slides for quantitative analysis.
           In Biomedical Imaging: From Nano to Macro, 2009.  ISBI'09.
           IEEE International Symposium on (pp. 1107-1110). IEEE.
    .. [#] Xu, J., Xiang, L., Wang, G., Ganesan, S., Feldman, M., Shih, N. N.,
           ...& Madabhushi, A. (2015). Sparse Non-negative Matrix Factorization
           (SNMF) based color unmixing for breast histopathological image
           analysis.  Computerized Medical Imaging and Graphics, 46, 20-29.

    """
    stains = ['hematoxylin', 'eosin'] if stains is None else stains
    stain_unmixing_routine_params = (
        {} if stain_unmixing_routine_params is None else
        stain_unmixing_routine_params)
    for k in ['W_source', 'mask_out']:
        assert k not in stain_unmixing_routine_params.keys(), \
            '%s must be provided as a separate parameter.' % k

    # find stains matrix from source image
    stain_unmixing_routine_params['stains'] = stains
    _, StainsFloat, _ = color_deconvolution_routine(
        im_src, W_source=W_source, mask_out=mask_out,
        **stain_unmixing_routine_params)

    # Get W_target

    if all(j is None for j in [W_target, im_target]):
        # Normalize to 'ideal' stain matrix if none is provided
        W_target = np.array(
            [stain_color_map[stains[0]], stain_color_map[stains[1]]]).T
        W_target = complement_stain_matrix(W_target)

    elif im_target is not None:
        # Get W_target from target image
        W_target = stain_unmixing_routine(
            im_target, **stain_unmixing_routine_params)

    # Convolve source image StainsFloat with W_target
    im_src_normalized = color_convolution(StainsFloat, W_target)

    # return masked values using unnormalized image
    if mask_out is not None:
        keep_mask = np.not_equal(mask_out, True)
        for i in range(3):
            original = im_src[:, :, i].copy()
            new = im_src_normalized[:, :, i].copy()
            original[keep_mask] = 0
            new[mask_out] = 0
            im_src_normalized[:, :, i] = new + original

    return im_src_normalized



import io
import os
import tempfile

import numpy as np
import pyvips
from imageio import imwrite
from PIL import Image

from histomicstk.annotations_and_masks.annotation_and_mask_utils import (
    get_image_from_htk_response, get_scale_factor_and_appendStr)
from histomicstk.annotations_and_masks.annotations_to_masks_handler import \
    _visualize_annotations_on_rgb
from histomicstk.annotations_and_masks.annotations_to_object_mask_handler import \
    get_all_rois_from_slide_v2
from histomicstk.annotations_and_masks.masks_to_annotations_handler import \
    get_annotation_documents_from_contours
from histomicstk.workflows.workflow_runner import (Slide_iterator,
                                                   Workflow_runner)

# CONSTANTS

# source: https://libvips.github.io/libvips/API/current/Examples.md.html
# source 2: https://libvips.github.io/libvips/API/current/Examples.md.html
# source 3: https://github.com/libvips/pyvips/issues/109
# source 4: https://github.com/libvips/libvips/issues/1254

# map np dtypes to vips
DTYPE_TO_FORMAT = {
    'uint8': 'uchar',
    'int8': 'char',
    'uint16': 'ushort',
    'int16': 'short',
    'uint32': 'uint',
    'int32': 'int',
    'float32': 'float',
    'float64': 'double',
    'complex64': 'complex',
    'complex128': 'dpcomplex',
}

# map vips formats to np dtypes
FORMAT_TO_DTYPE = {
    'uchar': np.uint8,
    'char': np.int8,
    'ushort': np.uint16,
    'short': np.int16,
    'uint': np.uint32,
    'int': np.int32,
    'float': np.float32,
    'double': np.float64,
    'complex': np.complex64,
    'dpcomplex': np.complex128,
}


def get_all_rois_from_folder_v2(
        gc, folderid, get_all_rois_kwargs, monitor=''):
    """Get all rois in a girder folder using get_all_rois_from_slide_v2().

    Parameters
    ----------
    gc : girder_client.Girder_Client
        authenticated girder client
    folderid : str
        girder id of folder
    get_all_rois_kwargs : dict
        kwargs to pass to get_all_rois_from_slide_v2()
    monitor : str
        monitor prefix

    Returns
    -------
    None

    """
    def _get_all_rois(slide_id, monitorPrefix, **kwargs):
        sld = gc.get('/item/%s' % slide_id)
        if '.' not in sld['name']:
            sld['name'] += '.'
        sldname = sld['name'][:sld['name'].find('.')].replace('/', '_#_')
        return get_all_rois_from_slide_v2(
            slide_id=slide_id, monitorprefix=monitorPrefix,
            # encoding slide id makes things easier later
            slide_name='%s_id-%s' % (sldname, slide_id),
            **kwargs)

    # update with params
    get_all_rois_kwargs['gc'] = gc

    # pull annotations for each slide in folder
    workflow_runner = Workflow_runner(
        slide_iterator=Slide_iterator(
            gc, source_folder_id=folderid,
            keep_slides=None,
        ),
        workflow=_get_all_rois,
        workflow_kwargs=get_all_rois_kwargs,
        monitorPrefix=monitor,
    )
    workflow_runner.run()


def _get_visualization_zoomout(
        gc, slide_id, bounds, MPP, MAG, zoomout=4):
    """Get a zoomed out visualization of ROI RGB and annotation overlay.

    Parameters
    ----------
    gc : girder_client.Girder_Client
        authenticated girder client
    slide_id : str
        girder ID of slide
    bounds : dict
        bounds of the region of interest. Must contain the keys
        XMIN, XMAX, YMIN, YMAX
    MPP : float
        Microns per pixel.
    MAG : float
        Magnification. MPP overrides this.
    zoomout : float
        how much to zoom out

    Returns
    -------
    np.array
        Zoomed out visualization. Output from _visualize_annotations_on_rgb().

    """
    # get append string for server request
    if MPP is not None:
        getsf_kwargs = {
            'MPP': MPP * (zoomout + 1),
            'MAG': None,
        }
    elif MAG is not None:
        getsf_kwargs = {
            'MPP': None,
            'MAG': MAG / (zoomout + 1),
        }
    else:
        getsf_kwargs = {
            'MPP': None,
            'MAG': None,
        }
    sf, appendStr = get_scale_factor_and_appendStr(
        gc=gc, slide_id=slide_id, **getsf_kwargs)

    # now get low-magnification surrounding field
    x_margin = (bounds['XMAX'] - bounds['XMIN']) * zoomout / 2
    y_margin = (bounds['YMAX'] - bounds['YMIN']) * zoomout / 2
    getStr = \
        '/item/%s/tiles/region?left=%d&right=%d&top=%d&bottom=%d' \
        % (slide_id,
           max(0, bounds['XMIN'] - x_margin),
           bounds['XMAX'] + x_margin,
           max(0, bounds['YMIN'] - y_margin),
           bounds['YMAX'] + y_margin)
    getStr += appendStr
    resp = gc.get(getStr, jsonResp=False)
    rgb_zoomout = get_image_from_htk_response(resp)

    # plot a bounding box at the ROI region
    xmin = x_margin * sf
    xmax = xmin + (bounds['XMAX'] - bounds['XMIN']) * sf
    ymin = y_margin * sf
    ymax = ymin + (bounds['YMAX'] - bounds['YMIN']) * sf
    xmin, xmax, ymin, ymax = (str(int(j)) for j in (xmin, xmax, ymin, ymax))
    contours_list = [{
        'color': 'rgb(255,255,0)',
        'coords_x': ','.join([xmin, xmax, xmax, xmin, xmin]),
        'coords_y': ','.join([ymin, ymin, ymax, ymax, ymin]),
    }]

    return _visualize_annotations_on_rgb(rgb_zoomout, contours_list)


def _get_review_visualization(rgb, vis, vis_zoomout):
    """Get a visualization of rgb and annotations for rapid review.

    Parameters
    ----------
    rgb : np.array
        mxnx3 rgb image
    vis : np.array
        visualization of rgb with overlaid annotations
    vis_zoomout
        same as vis, but at a lower magnififcation.

    Returns
    -------
    np.array
        visualization to be used for gallery

    """
    import matplotlib.pyplot as plt

    wmax = max(vis.shape[1], vis_zoomout.shape[1])
    hmax = max(vis.shape[0], vis_zoomout.shape[0])

    fig, ax = plt.subplots(
        1, 3, dpi=100,
        figsize=(3 * wmax / 1000, hmax / 1000),
        gridspec_kw={'wspace': 0.01, 'hspace': 0},
    )

    ax[0].imshow(vis)
    ax[1].imshow(rgb)
    ax[2].imshow(vis_zoomout)

    for axis in ax:
        axis.axis('off')
    fig.subplots_adjust(bottom=0, top=1, left=0, right=1)

    buf = io.BytesIO()
    plt.savefig(buf, format='png', pad_inches=0, dpi=1000)
    buf.seek(0)
    combined_vis = np.uint8(Image.open(buf))[..., :3]
    plt.close()

    return combined_vis


def _plot_rapid_review_vis(
        roi_out, gc, slide_id, slide_name, MPP, MAG,
        combinedvis_savepath, zoomout=4,
        verbose=False, monitorprefix=''):
    """Plot a visualization for rapid review of ROI.

    This is a callback to be called inside get_all_rois_from_slide_v2().

    Parameters
    ----------
    roi_out : dict
        output from annotations_to_contours_no_mask()
    gc : girder_client.Girder_Client
        authenticated girder client
    slide_id : str
        girder slide id
    slide_name : str
        name of the slide
    MPP : float
        microns per pixel
    MAG : float
        magnification. superseded by MPP.
    combinedvis_savepath : str
        path to save the combined visualization
    zoomout : float
        how much to zoom out to get the gallery visualization
    verbose : bool
        print statements to screen
    monitorprefix : str
        text to prepent to printed statements

    Returns
    -------
    dict
        roi_out parameter whether or not it is modified

    """
    # get rgb and visualization (fetched mag + lower mag)
    vis_zoomout = _get_visualization_zoomout(
        gc=gc, slide_id=slide_id, bounds=roi_out['bounds'],
        MPP=MPP, MAG=MAG, zoomout=zoomout)

    # combined everything in a neat visualization for rapid review
    ROINAMESTR = '%s_left-%d_top-%d_bottom-%d_right-%d' % (
        slide_name,
        roi_out['bounds']['XMIN'], roi_out['bounds']['YMIN'],
        roi_out['bounds']['YMAX'], roi_out['bounds']['XMAX'])
    savename = os.path.join(combinedvis_savepath, ROINAMESTR + '.png')
    rapid_review_vis = _get_review_visualization(
        rgb=roi_out['rgb'], vis=roi_out['visualization'],
        vis_zoomout=vis_zoomout)

    # save visualization for later use
    if verbose:
        print('%s: Saving %s' % (monitorprefix, savename))
    imwrite(im=rapid_review_vis, uri=savename)

    return roi_out


def create_review_galleries(
        tilepath_base, upload_results=True, gc=None,
        gallery_savepath=None, gallery_folderid=None,
        padding=25, tiles_per_row=2, tiles_per_column=5,
        annprops=None, url=None, nameprefix=''):
    """Create and or post review galleries for rapid review.

    Parameters
    ----------
    tilepath_base : str
        directory where combined visualization.
    upload_results : bool
        upload results to DSA?
    gc : girder_client.Girder_Client
        authenticated girder client. Only needed upload_results.
    gallery_savepath : str
        directory to save gallery. Only if upload_results.
    gallery_folderid : str
        girder ID of folder to post galleries. Only if upload_result.
    padding : int
        padding in pixels between tiles in same gallery.
    tiles_per_row : int
        how many visualization tiles per row in gallery.
    tiles_per_column : int
        how many visualization tiles per column in gallery.
    annprops : dict
        properties of the annotations to be posted to DSA. Passed directly
        as annprops to get_annotation_documents_from_contours()
    url : str
        url of the Digital Slide Archive Instance. For example:
        http://candygram.neurology.emory.edu:8080/
    nameprefix : str
        prefix to prepend to gallery name

    Returns
    -------
    list
        each entry is a dict representing the response of the server
        post request to upload the gallery to DSA.

    """
    from pandas import DataFrame

    if upload_results:
        for par in ('gc', 'gallery_folderid', 'url'):
            if locals()[par] is None:
                raise Exception(
                    '%s cannot be None if upload_results!' % par)

    if gallery_savepath is None:
        gallery_savepath = tempfile.mkdtemp(prefix='gallery-')

    savepaths = []
    resps = []

    tile_paths = sorted([
        os.path.join(tilepath_base, j) for j in
        os.listdir(tilepath_base) if j.endswith('.png')])

    def _parse_tilepath(tpath):
        basename = os.path.basename(tpath)
        basename = basename[:basename.rfind('.')]
        tileinfo = {'slide_name': basename.split('_')[0]}
        for attrib in ['id', 'left', 'top', 'bottom', 'right']:
            tileinfo[attrib] = basename.split(
                attrib + '-')[1].split('_')[0]

        # add URL in histomicsTK
        tileinfo['URL'] = url + \
            'histomicstk#?image=%s&bounds=%s%%2C%s%%2C%s%%2C%s%%2C0' % (
                tileinfo['id'],
                tileinfo['left'], tileinfo['top'],
                tileinfo['right'], tileinfo['bottom'])
        return tileinfo

    n_tiles = len(tile_paths)
    n_galleries = int(np.ceil(n_tiles / (tiles_per_row * tiles_per_column)))

    tileidx = 0

    for galno in range(n_galleries):

        # this makes a 8-bit, mono image (initializes as 1x1x3 matrix)
        im = pyvips.Image.black(1, 1, bands=3)

        # this will store the roi contours
        contours = []

        for _row in range(tiles_per_column):

            rowpos = im.height + padding

            # initialize "row" strip image
            row_im = pyvips.Image.black(1, 1, bands=3)

            for _col in range(tiles_per_row):

                if tileidx == n_tiles:
                    break

                tilepath = tile_paths[tileidx]

                print('Inserting tile %d of %d: %s' % (
                    tileidx, n_tiles, tilepath))
                tileidx += 1

                # # get tile from file
                tile = pyvips.Image.new_from_file(
                    tilepath, access='sequential')

                # insert tile into mosaic row
                colpos = row_im.width + padding
                row_im = row_im.insert(
                    tile[:3], colpos, 0, expand=True, background=255)

                if upload_results:

                    tileinfo = _parse_tilepath(tilepath)

                    xmin = colpos
                    ymin = rowpos
                    xmax = xmin + tile.width
                    ymax = ymin + tile.height
                    xmin, xmax, ymin, ymax = (
                        str(j) for j in (xmin, xmax, ymin, ymax))
                    contours.append({
                        'group': tileinfo['slide_name'],
                        'label': tileinfo['URL'],
                        'color': 'rgb(0,0,0)',
                        'coords_x': ','.join([xmin, xmax, xmax, xmin, xmin]),
                        'coords_y': ','.join([ymin, ymin, ymax, ymax, ymin]),
                    })

                    # Add a small contour so that when the pathologist
                    # changes the label to approve or disapprove of the
                    # FOV, the URL in THIS contour (a link to the original
                    # FOV) can be used. We place it in the top right corner.
                    boxsize = 25
                    xmin = str(int(xmax) - boxsize)
                    ymax = str(int(ymin) + boxsize)
                    contours.append({
                        'group': tileinfo['slide_name'],
                        'label': tileinfo['URL'],
                        'color': 'rgb(0,0,0)',
                        'coords_x': ','.join([xmin, xmax, xmax, xmin, xmin]),
                        'coords_y': ','.join([ymin, ymin, ymax, ymax, ymin]),
                    })

            # insert row into main gallery
            im = im.insert(row_im, 0, rowpos, expand=True, background=255)

        filename = '%s_gallery-%d' % (nameprefix, galno + 1)
        savepath = os.path.join(gallery_savepath, filename + '.tiff')
        print('Saving gallery %d of %d to %s' % (
            galno + 1, n_galleries, savepath))

        # save temporarily to disk to be uploaded
        im.tiffsave(
            savepath, tile=True, tile_width=256, tile_height=256, pyramid=True)

        if upload_results:
            # upload the gallery to DSA
            resps.append(gc.uploadFileToFolder(
                folderId=gallery_folderid, filepath=savepath,
                filename=filename))
            os.remove(savepath)

            # get and post FOV location annotations
            annotation_docs = get_annotation_documents_from_contours(
                DataFrame(contours), separate_docs_by_group=True,
                annprops=annprops)
            for doc in annotation_docs:
                _ = gc.post(
                    '/annotation?itemId=' + resps[-1]['itemId'], json=doc)
        else:
            savepaths.append(savepath)

    return resps if upload_results else savepaths



"""Compute intensity features in labeled image."""
import warnings

import numpy as np


def compute_intensity_features(
        im_label, im_intensity, num_hist_bins=10,
        rprops=None, feature_list=None):
    """Calculate intensity features from an intensity image.

    Parameters
    ----------
    im_label : array_like
        A labeled mask image wherein intensity of a pixel is the ID of the
        object it belongs to. Non-zero values are considered to be foreground
        objects.

    im_intensity : array_like
        Intensity image.

    num_hist_bins: int, optional
        Number of bins used to computed the intensity histogram of an object.
        Histogram is used to energy and entropy features. Default is 10.

    rprops : output of skimage.measure.regionprops, optional
        rprops = skimage.measure.regionprops( im_label ). If rprops is not
        passed then it will be computed inside which will increase the
        computation time.

    feature_list : list, default is None
        list of intensity features to return.
        If none, all intensity features are returned.

    Returns
    -------
    fdata: pandas.DataFrame
        A pandas dataframe containing the intensity features listed below for
        each object/label.

    Notes
    -----
    List of intensity features computed by this function:

    Intensity.Min : float
        Minimum intensity of object pixels.

    Intensity.Max : float
        Maximum intensity of object pixels.

    Intensity.Mean : float
        Mean intensity of object pixels

    Intensity.Median : float
        Median intensity of object pixels

    Intensity.MeanMedianDiff : float
        Difference between mean and median intensities of object pixels.

    Intensity.Std : float
        Standard deviation of the intensities of object pixels

    Intensity.IQR: float
        Inter-quartile range of the intensities of object pixels

    Intensity.MAD: float
        Median absolute deviation of the intensities of object pixels

    Intensity.Skewness : float
        Skewness of the intensities of object pixels. Value is 0 when all
        intensity values are equal.

    Intensity.Kurtosis : float
        Kurtosis of the intensities of object pixels. Value is -3 when all
        values are equal.

    Intensity.HistEnergy : float
        Energy of the intensity histogram of object pixels

    Intensity.HistEntropy : float
        Entropy of the intensity histogram of object pixels.

    References
    ----------
    .. [#] Daniel Zwillinger and Stephen Kokoska. "CRC standard probability
       and statistics tables and formulae," Crc Press, 1999.

    """
    import pandas as pd
    import scipy.stats
    from skimage.measure import regionprops

    default_feature_list = [
        'Intensity.Min',
        'Intensity.Max',
        'Intensity.Mean',
        'Intensity.Median',
        'Intensity.MeanMedianDiff',
        'Intensity.Std',
        'Intensity.IQR',
        'Intensity.MAD',
        'Intensity.Skewness',
        'Intensity.Kurtosis',
        'Intensity.HistEnergy',
        'Intensity.HistEntropy',
    ]

    # List of feature names
    if feature_list is None:
        feature_list = default_feature_list
    else:
        assert all(j in default_feature_list for j in feature_list), \
            'Some feature names are not recognized.'

    # compute object properties if not provided
    if rprops is None:
        rprops = regionprops(im_label)

    # create pandas data frame containing the features for each object
    numFeatures = len(feature_list)
    numLabels = len(rprops)
    fdata = pd.DataFrame(np.zeros((numLabels, numFeatures)),
                         columns=feature_list)

    # conditionally execute calculations if x in the features list
    def _conditional_execution(feature, func, *args, **kwargs):
        if feature in feature_list:
            with warnings.catch_warnings():
                warnings.simplefilter('ignore', RuntimeWarning)
                fdata.at[i, feature] = func(*args, **kwargs)

    def _return_input(x):
        return x

    for i in range(numLabels):
        if rprops[i] is None:
            continue

        # get intensities of object pixels
        pixelIntensities = np.sort(
            im_intensity[rprops[i].coords[:, 0], rprops[i].coords[:, 1]],
        )

        # simple descriptors
        meanIntensity = np.mean(pixelIntensities)
        medianIntensity = np.median(pixelIntensities)
        _conditional_execution('Intensity.Min', np.min, pixelIntensities)
        _conditional_execution('Intensity.Max', np.max, pixelIntensities)
        _conditional_execution('Intensity.Mean', _return_input, meanIntensity)
        _conditional_execution(
            'Intensity.Median', _return_input, medianIntensity)
        _conditional_execution(
            'Intensity.MeanMedianDiff', _return_input,
            meanIntensity - medianIntensity)
        _conditional_execution('Intensity.Std', np.std, pixelIntensities)
        _conditional_execution(
            'Intensity.Skewness', scipy.stats.skew, pixelIntensities)
        _conditional_execution(
            'Intensity.Kurtosis', scipy.stats.kurtosis, pixelIntensities)

        # inter-quartile range
        _conditional_execution(
            'Intensity.IQR', scipy.stats.iqr, pixelIntensities)

        # median absolute deviation
        _conditional_execution(
            'Intensity.MAD', np.median,
            np.abs(pixelIntensities - medianIntensity))

        # histogram-based features
        if any(j in feature_list for j in [
                'Intensity.HistEntropy', 'Intensity.HistEnergy']):

            # compute intensity histogram
            hist, bins = np.histogram(pixelIntensities, bins=num_hist_bins)
            prob = hist / np.sum(hist, dtype=np.float32)

            # entropy and energy
            _conditional_execution(
                'Intensity.HistEntropy', scipy.stats.entropy, prob)
            _conditional_execution('Intensity.HistEnergy', np.sum, prob**2)

    return fdata



import numpy as np

cimport cython
cimport numpy as np


@cython.cdivision(True)
@cython.boundscheck(False)
@cython.wraparound(False)
def _compute_marginal_glcm_probs_cython(double[:, :] nGLCM not None):

    cdef long num_levels = nGLCM.shape[0]

    cdef double[::1] px = np.zeros([num_levels], dtype=float)
    cdef double[::1] py = np.zeros([num_levels], dtype=float)

    cdef double[::1] pxPlusy = np.zeros([2*num_levels-1], dtype=float)
    cdef double[::1] pxMinusy = np.zeros([num_levels], dtype=float)
    cdef int i, j, i_minus_j

    with nogil:
        for i in range(num_levels):
            for j in range(num_levels):
                px[i] = px[i] + nGLCM[i, j]
                py[i]= py[i] + nGLCM[j, i]
                pxPlusy[i+j] = pxPlusy[i+j] + nGLCM[i, j]
                i_minus_j = i-j
                if i < j:
                    i_minus_j *= -1
                pxMinusy[i_minus_j] = pxMinusy[i_minus_j] + nGLCM[i, j]

    return np.asarray(px), np.asarray(py), np.asarray(pxPlusy), np.asarray(pxMinusy)



import numpy as np


def glog(im_input, alpha=1, range=None, theta=np.pi / 4, tau=0.6, eps=0.6):
    """Performs generalized Laplacian of Gaussian blob detection.

    Parameters
    ----------
    im_input : array_like
        A hematoxylin intensity image obtained from ColorDeconvolution.
    alpha : double
        A positive scalar used to normalize the gLoG filter responses. Controls
        the blob-center detection and eccentricities of detected blobs. Larger
        values emphasize more eccentric blobs. Default value = 1.
    range : array_like
        Scale range
    theta : double
        Angular increment for rotating gLoG filters. Default value = np.pi / 6.
    tau : double
        Tolerance for counting pixels in determining optimal scale SigmaC
    eps : double
        range to define SigmaX surrounding SigmaC

    Returns
    -------
    Rsum : array_like
        Sum of filter responses at specified scales and orientations
    Maxima: : array_like
        A binary mask highlighting maxima pixels

    Notes
    -----
    Return values are returned as a namedtuple

    References
    ----------
    .. [#] H. Kong, H.C. Akakin, S.E. Sarma, "A Generalized Laplacian of
       Gaussian Filter for Blob Detection and Its Applications," in IEEE
       Transactions on Cybernetics, vol.43,no.6,pp.1719-33, 2013.

    """
    import scipy.ndimage as ndi
    from skimage import morphology

    range = np.linspace(1.5, 3, int(np.round((3 - 1.5) / 0.2)) + 1) if range is None else range

    # initialize sigma
    Sigma = np.exp(range)

    # generate circular LoG scale-space to determine range of SigmaX
    l_g = 0
    H = []
    Bins = []
    Min = np.zeros((len(Sigma), 1))
    Max = np.zeros((len(Sigma), 1))
    for i, s in enumerate(Sigma):
        Response = s**2 * ndi.gaussian_laplace(im_input, s, output=None,
                                               mode='constant',
                                               cval=0.0)
        Min[i] = Response.min()
        Max[i] = Response.max()
        Bins.append(np.arange(0.01 * np.floor(Min[i] / 0.01),
                    0.01 * np.ceil(Max[i] / 0.01) + 0.01, 0.01))
        Hist = np.histogram(Response, Bins[i])
        H.append(Hist[0])
        if Max[i] > l_g:
            l_g = Max[i]

    # re-normalized based on global max and local min, count threshold pixels
    Zeta = np.zeros((len(Sigma), 1))
    for i, _s in enumerate(Sigma):
        Bins[i] = (Bins[i] - Min[i]) / (l_g - Min[i])
        Zeta[i] = np.sum(H[i][Bins[i][0:-1] > tau])

    # identify best scale SigmaC based on maximum circular response
    Index = np.argmax(Zeta)

    # define range for SigmaX
    XRange = range(max(Index - 2, 0), min(len(range), Index + 2) + 1)
    SigmaX = np.exp(range[XRange])

    # define rotation angles
    Thetas = np.linspace(0, np.pi - theta, int(np.round(np.pi / theta)))

    # loop over SigmaX, SigmaY and then angle, summing up filter responses
    Rsum = np.zeros(im_input.shape)
    for i, Sx in enumerate(SigmaX):
        YRange = range(0, XRange[i])
        SigmaY = np.exp(range[YRange])
        for Sy in SigmaY:
            for Th in Thetas:
                Kernel = glogkernel(Sx, Sy, Th)
                Kernel *= (1 + np.log(Sx) ** alpha) * (1 + np.log(Sy) ** alpha)
                Rsum += ndi.convolve(im_input, Kernel, mode='constant', cval=0.0)
                print(Sx, Sy, Th)
        Kernel = glogkernel(Sx, Sx, 0)
        Kernel *= (1 + np.log(Sx) ** alpha) * (1 + np.log(Sx) ** alpha)
        Rsum += ndi.convolve(im_input, Kernel, mode='constant', cval=0.0)
        print(Sx, Sx, 0)

    # detect local maxima
    Disk = morphology.disk(3 * np.exp(range[Index]))
    Maxima = ndi.maximum_filter(Rsum, footprint=Disk)
    Maxima = Rsum == Maxima

    return Rsum, Maxima


def glogkernel(sigma_x, sigma_y, theta):

    N = np.ceil(2 * 3 * sigma_x)
    X, Y = np.meshgrid(np.linspace(0, N, int(N + 1)) - N / 2,
                       np.linspace(0, N, int(N + 1)) - N / 2)
    a = np.cos(theta) ** 2 / (2 * sigma_x ** 2) + \
        np.sin(theta) ** 2 / (2 * sigma_y ** 2)
    b = -np.sin(2 * theta) / (4 * sigma_x ** 2) + \
        np.sin(2 * theta) / (4 * sigma_y ** 2)
    c = np.sin(theta) ** 2 / (2 * sigma_x ** 2) + \
        np.cos(theta) ** 2 / (2 * sigma_y ** 2)
    D2Gxx = ((2 * a * X + 2 * b * Y)**2 - 2 * a) * np.exp(-(a * X**2 + 2 * b * X * Y + c * Y**2))
    D2Gyy = ((2 * b * X + 2 * c * Y)**2 - 2 * c) * np.exp(-(a * X**2 + 2 * b * X * Y + c * Y**2))
    Gaussian = np.exp(-(a * X**2 + 2 * b * X * Y + c * Y**2))
    Kernel = (D2Gxx + D2Gyy) / np.sum(Gaussian.flatten())
    return Kernel



import numpy as np


def gradient_diffusion(im_dx, im_dy, im_fgnd_mask,
                       mu=5, lambda_=5, iterations=10, dt=0.05):
    """
    Diffusion of gradient field using Navier-Stokes equation. Used for
    smoothing/denoising a gradient field.

    Takes as input a gradient field image (dX, dY), and a mask of the
    foreground region, and then iteratively solves the Navier-Stokes equation
    to diffuse the vector field and align noisy gradient vectors with their
    surrounding signals.

    Parameters
    ----------
    im_dx : array_like
        Horizontal component of gradient image.
    im_dy : array_like
        Vertical component of gradient image.
    im_fgnd_mask : array_like
        Binary mask where foreground objects have value 1, and background
        objects have value 0. Used to restrict influence of background vectors
        on diffusion process.
    mu : float
        Weight parameter from Navier-Stokes equation - weights divergence and
        Laplacian terms. Default value = 5.
    lambda_ : float
        Weight parameter from Navier-Stokes equation - used to weight
        divergence. Default value = 5.
    iterations : float
        Number of time-steps to use in solving Navier-Stokes. Default value =
        10.
    dt : float
        Timestep to be used in solving Navier-Stokes. Default value = 0.05.

    Returns
    -------
    im_vx : array_like
        Horizontal component of diffused gradient.
    im_vy : array_like
        Vertical component of diffused gradient.

    See Also
    --------
    histomicstk.segmentation.nuclear.GradientFlow

    References
    ----------
    .. [#] G. Li et al "3D cell nuclei segmentation based on gradient flow
           tracking" in BMC Cell Biology,vol.40,no.8, 2007.

    """
    import scipy.ndimage as ndi

    # initialize solution
    im_vx = im_dx.copy()
    im_vy = im_dy.copy()

    # iterate for prescribed number of iterations
    for _it in range(iterations):

        # calculate divergence of current solution
        vXY, vXX = np.gradient(im_vx)
        vYY, vYX = np.gradient(im_vy)
        Div = vXX + vYY
        DivY, DivX = np.gradient(Div)

        # calculate laplacians of current solution
        im_vx += dt * (mu * ndi.laplace(im_vx) +
                       (lambda_ + mu) * DivX +
                       im_fgnd_mask * (im_dx - im_vx))
        im_vy += dt * (mu * ndi.laplace(im_vy) +
                       (lambda_ + mu) * DivY +
                       im_fgnd_mask * (im_dy - im_vy))

    # return solution
    return im_vx, im_vy



"""
This package contains functions for enhancing different kinds of structures
(e.g. edges/membrane, blobs/nuclei, vessels) in images.
"""
# import sub-packages to support nested calls
from . import edge, shape

# list out things that are available for public use
__all__ = (
    # sub-packages
    'edge',
    'shape',
)



import json
from pathlib import Path

import numpy

import histomicstk
import histomicstk.preprocessing.color_deconvolution as htk_cdeconv
from histomicstk.cli import utils
from histomicstk.cli.utils import CLIArgumentParser


def main(origargs):
    args = utils.splitArgs(origargs)
    args.snmf.I_0 = numpy.array(args.snmf.I_0)

    print('>> Starting Dask cluster and sampling pixels')
    utils.create_dask_client(args.dask)
    sample = utils.sample_pixels(args.sample)

    # Create stain matrix
    print('>> Creating stain matrix')

    args.snmf.w_init = utils.get_stain_matrix(args.stains, 2)

    print(args.snmf.w_init)

    # Perform color deconvolution
    print('>> Performing color deconvolution')

    w_est = htk_cdeconv.rgb_separate_stains_xu_snmf(sample.T, **vars(args.snmf))
    w_est = htk_cdeconv.complement_stain_matrix(w_est)

    annotation = {
        'name': 'SeperateStainsXuSnmf',
        'attributes': {
            'params': vars(origargs),
            'cli': Path(__file__).stem,
            'version': histomicstk.__version__,
        },
    }
    for i, stain in enumerate(w_est.T):
        annotation['attributes'][f'stainColor_{i + 1}'] = stain.tolist()
    with open(args.outputAnnotationFile, 'w') as annotation_file:
        json.dump(annotation, annotation_file, sort_keys=False)


if __name__ == '__main__':
    main(CLIArgumentParser().parse_args())



"""This module contains various tools to detect tissue.

Included functionality includes methods to detect tissue in a whole-slide image
(i.e. segment foreground from white space), as well as specific workflows
that detect artifacts, specific components (eg blood), and
highly-cellular regions within detected tissue.

"""



"""Placeholder."""
import numpy
import numpy.linalg as np_linalg

import histomicstk.utils as utils

from . import _linalg as htk_linalg


def separate_stains_xu_snmf(im_sda, w_init=None, beta=0.2):
    """Compute the stain matrix for color deconvolution with SNMF.

    ... (sparse non-negative matrix factorization).

    Parameters
    ----------
    im_sda : array_like
        Image (MxNx3) or matrix (3xN) in SDA space for which to compute the
        stain matrix.
    w_init : array_like, default is None
        Initial value for the stain matrix.  if not provided, default
        initialization is used.
    beta : float
        Regularization factor for the sparsity of the deconvolved pixels

    Returns
    -------
    w : array_like
        A 3x3 matrix of stain column vectors

    Note
    ----
    All input pixels are used in the factorization.

    See Also
    --------
    histomicstk.preprocessing.color_deconvolution.color_deconvolution
    histomicstk.preprocessing.color_deconvolution.separate_stains_macenko_pca

    References
    ----------
    .. [#] Van Eycke, Y. R., Allard, J., Salmon, I., Debeir, O., &
           Decaestecker, C. (2017).  Image processing in digital pathology: an
           opportunity to solve inter-batch variability of immunohistochemical
           staining.  Scientific Reports, 7.
    .. [#] Xu, J., Xiang, L., Wang, G., Ganesan, S., Feldman, M., Shih, N. N.,
           ... & Madabhushi, A. (2015). Sparse Non-negative Matrix Factorization
           (SNMF) based color unmixing for breast histopathological image
           analysis.  Computerized Medical Imaging and Graphics, 46, 20-29.

    """
    import nimfa

    # Image matrix
    m = utils.convert_image_to_matrix(im_sda)
    m = utils.exclude_nonfinite(m)
    factorization = \
        nimfa.Snmf(m, rank=m.shape[0] if w_init is None else w_init.shape[1],
                   W=w_init,
                   H=None if w_init is None else np_linalg.pinv(w_init).dot(m),
                   beta=beta)
    factorization.factorize()
    return htk_linalg.normalize(numpy.array(factorization.W))



import colorsys
import json
import os
from pathlib import Path

import dask
import dask.dataframe as dd
import large_image
import numpy as np
import pandas as pd

import histomicstk.segmentation.label as htk_seg_label
import histomicstk.segmentation.nuclear as htk_nuclear

try:
    import joblib
except ImportError:
    # Versions of scikit-learn before 0.21 had joblib internally
    from sklearn.externals import joblib

import logging

import histomicstk
from histomicstk.cli import utils as cli_utils
from histomicstk.cli.utils import CLIArgumentParser

logging.basicConfig(level=logging.CRITICAL)


def set_reference_values(args):
    """
    Set reference values and configuration parameters for feature extraction.

    Args:
    ----
        args (dict): Configuration parameters for feature extraction.

    Returns:
    -------
        dict: Updated configuration parameters with reference values set.

    """
    args.reference_mu_lab = [8.63234435, -0.11501964, 0.03868433]
    args.reference_std_lab = [0.57506023, 0.10403329, 0.01364062]
    args.foreground_threshold = 60
    args.min_radius = 6
    args.max_radius = 20
    args.min_nucleus_area = 80
    args.local_max_search_radius = 10
    args.nuclei_annotation_format = 'boundary'
    args.stain_1 = 'hematoxylin'
    args.stain_1_vector = [-1.0, -1.0, -1.0]
    args.stain_2 = 'eosin'
    args.stain_2_vector = [-1.0, -1.0, -1.0]
    args.stain_3 = 'null'
    args.stain_3_vector = [-1.0, -1.0, -1.0]
    args.ignore_border_nuclei = False
    args.cyto_width = 8
    args.cytoplasm_features = True
    args.fsd_bnd_pts = 128
    args.fsd_features = True
    args.fsd_freq_bins = 6
    args.gradient_features = True
    args.haralick_features = True
    args.morphometry_features = True
    args.intensity_features = True
    args.gradient_features = True
    args.fsd_features = True
    args.num_glcm_levels = 32
    args.min_fgnd_frac = .25
    args.analysis_roi = None
    return args


def gen_distinct_rgb_colors(n, seed=None):
    """
    Generates N visually distinct RGB colors

    Parameters
    ----------
    n : int
        Number of distinct RGB colors to output

    seed: int or array_like, optional
        Seed for the random number generator. See documentation of
        `numpy.random.seed` for more details

    Returns
    -------
    colors_list : list
        A list of n RGB colors

    References
    ----------
    .. [#] http://martin.ankerl.com/2009/12/09/how-to-create-random-colors-programmatically/  # noqa

    """
    np.random.seed(seed)
    h = np.random.random()
    np.random.seed(None)

    golden_ratio_conjugate = (np.sqrt(5) - 1) / 2.0

    color_list = [colorsys.hsv_to_rgb((h + i * golden_ratio_conjugate) % 1,
                                      1.0, 1.0)
                  for i in range(n)]

    return color_list


def process_feature_and_annotation(args):
    """
    Process nuclei feature extraction and annotation from an input image.

    Args:
    ----
        args (dict): Configuration parameters for feature extraction.

    Returns:
    -------
        tuple: A tuple containing nuclei annotations (list) and feature data (Dask DataFrame).

    """
    print('>> Generating features and annotation')

    #
    # Set arguments required for nuclei feature extraction
    #
    args = set_reference_values(args)
    tile_overlap = (args.max_radius + 1) * 4
    it_kwargs = {'tile_overlap': {'x': tile_overlap, 'y': tile_overlap}}

    #
    # Read Input Image
    #
    print('\n>> Reading input image ... \n')

    ts = large_image.getTileSource(args.inputImageFile)

    ts_metadata = ts.getMetadata()

    print(json.dumps(ts_metadata, indent=2))

    src_mu_lab = None
    src_sigma_lab = None

    #
    # Detect and compute nuclei features in parallel using Dask
    #
    print('\n>> Detecting nuclei and computing features ...\n')

    tile_result_list = []

    for tile in ts.tileIterator(**it_kwargs):

        # detect nuclei
        cur_result = dask.delayed(htk_nuclear.detect_tile_nuclei)(
            tile,
            args,
            src_mu_lab, src_sigma_lab,
            return_fdata=True,
        )

        # append result to list
        tile_result_list.append(cur_result)

    tile_result_list = dask.delayed(tile_result_list).compute()

    nuclei_annot_list = [annot
                         for annot_list, fdata in tile_result_list
                         for annot in annot_list]

    # remove overlapping nuclei
    nuclei_annot_list = htk_seg_label.remove_overlap_nuclei(
        nuclei_annot_list, args.nuclei_annotation_format)

    nuclei_fdata = pd.DataFrame()

    if len(nuclei_annot_list) > 0:

        nuclei_fdata = pd.concat([
            fdata
            for annot_list, fdata in tile_result_list if fdata is not None],
            ignore_index=True,
        )
    # Fill any instances with NaN as zero
    df = pd.DataFrame(nuclei_fdata).fillna(0)
    return nuclei_annot_list, dd.from_pandas(df, npartitions=1)


def read_feature_file(args):
    """
    Read nuclei feature data from a specified file.

    Args:
    ----
        args (dict): Configuration parameters including the input feature file path.

    Returns:
    -------
        dask.dataframe.DataFrame: A Dask DataFrame containing the nuclei feature data.

    """
    fname, feature_file_format = os.path.splitext(args.inputNucleiFeatureFile)

    if feature_file_format == '.csv':

        ddf = dd.read_csv(args.inputNucleiFeatureFile)

    elif feature_file_format == '.h5':

        ddf = dd.read_hdf(args.inputNucleiFeatureFile, 'Features')

    else:
        msg = 'Extension of output feature file must be .csv or .h5'
        raise ValueError(msg)

    # Fill any instances with NaN as zero
    return ddf.fillna(0)


def main(args):

    print('\n>> CLI Parameters ...\n')

    print(args)

    #
    # Initiate Dask client
    #
    print('\n>> Creating Dask client ...\n')

    c = cli_utils.create_dask_client(args)

    print(c)

    #
    # read model file
    #
    print('\n>> Loading classification model ...\n')
    clf_model = joblib.load(args.inputModelFile)

    if args.inputNucleiFeatureFile and args.inputNucleiAnnotationFile:

        # read feature file
        print('\n>> Loading nuclei feature file ...\n')

        ddf = read_feature_file(args)

        if len(ddf.columns) != clf_model.n_features_in_:

            msg = (
                'The number of features of the classification model '
                'and the input feature file do not match.'
            )
            raise ValueError(msg)

        #
        # read nuclei annotation file
        #
        print('\n>> Loading nuclei annotation file ...\n')

        with open(args.inputNucleiAnnotationFile) as f:

            annotation_data = json.load(f)
            nuclei_annot_list = annotation_data.get(
                'elements', annotation_data.get(
                    'annotation', {}).get('elements'))

        if len(nuclei_annot_list) != len(ddf.index):

            msg = (
                'The number of nuclei in the feature file and the '
                'annotation file do not match'
            )
            raise ValueError(msg)
    else:
        nuclei_annot_list, ddf = process_feature_and_annotation(args)

    #
    # Perform nuclei classification
    #
    print('\n>> Performing nuclei classification using Dask ...\n')

    def predict_nuclei_class_prob(df, clf_model):

        return pd.DataFrame(data=clf_model.predict_proba(df.values),
                            columns=clf_model.classes_)

    outfmt = pd.DataFrame(columns=clf_model.classes_, dtype=np.float64)

    df_class_prob = ddf.map_partitions(predict_nuclei_class_prob, clf_model,
                                       meta=outfmt).compute()

    pred_class = df_class_prob.idxmax(axis=1)

    #
    # Group nuclei annotations by class
    #
    print('\n>> Grouping nuclei annotations by class ...\n')

    num_classes = len(clf_model.classes_)

    nuclei_annot_by_class = {c: [] for c in clf_model.classes_}

    class_color_map = dict(zip(clf_model.classes_,
                               gen_distinct_rgb_colors(num_classes, seed=1)))

    for i in range(len(nuclei_annot_list)):

        cur_class = pred_class.iloc[i]

        cur_anot = nuclei_annot_list[i]
        cur_anot['lineColor'] = 'rgb(%s)' % ','.join(
            [str(int(round(col * 255))) for col in class_color_map[cur_class]])
        nuclei_annot_by_class[cur_class].append(cur_anot)

    #
    # Write annotation file
    #
    print('\n>> Writing classified nuclei annotation file ...\n')

    annot_fname = os.path.splitext(
        os.path.basename(args.outputNucleiAnnotationFile))[0]

    annotation = []
    for c in clf_model.classes_:

        annotation.append({
            'name': 'Class ' + str(c) + ' ' + annot_fname,
            'elements': nuclei_annot_by_class[c],
        })
    annotation.append({
        'attributes': {
            'params': vars(args),
            'cli': Path(__file__).stem,
            'version': histomicstk.__version__,

        },
    })
    with open(args.outputNucleiAnnotationFile, 'w') as annotation_file:
        json.dump(annotation, annotation_file, separators=(',', ':'), sort_keys=False)


if __name__ == '__main__':
    main(CLIArgumentParser().parse_args())
